#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import List
from mitosheet.types import StepsManagerType

UNDO_EVENT = 'undo'
UNDO_PARAMS: List[str] = []

def execute_undo_update(steps_manager: StepsManagerType) -> None:
    steps_manager.execute_undo()

"""
This object wraps all the information
that is needed for a undo step!
"""
UNDO_UPDATE = {
    'event_type': UNDO_EVENT,
    'params': UNDO_PARAMS,
    'execute': execute_undo_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import re
import pandas as pd
from typing import Any, Dict
from mitosheet.types import StepsManagerType


def get_search_matches(params: Dict[str, Any], steps_manager: StepsManagerType) -> Any:
    """
    Finds the number of matches to a given search value in the dataframe.
    """
    sheet_index = params['sheet_index']
    search_value = params['search_value']
    df = steps_manager.dfs[sheet_index]

    # First, generate a count for each value in the dataframe:
    unique_value_counts = df.stack().value_counts()

    escaped_search_value = re.escape(search_value)

    # Use the same regex for all searching
    search_regex = re.compile(escaped_search_value, re.IGNORECASE)

    # Then, filter for the search value:
    matches = unique_value_counts.filter(regex=search_regex, axis=0)

    # Then, sum the matches:
    total_number_matches = int(matches.sum())

    # Then, add the column names to the matches:
    total_number_matches += len([col for col in df.columns if re.search(search_regex,str(col)) is not None])

    # Get the number of rows and columns so that we don't have to calculate the length in each iteration
    num_rows = len(df.index)
    num_cols = len(df.columns)

    # Find the indices of cells containing the search value
    # Only search the first 1500 rows because the editor only shows the first 1500 rows. 
    cell_matches = [{'rowIndex': i, 'colIndex': j} for i in range(min(1500, num_rows)) for j in range(num_cols) if (re.search(search_regex,str(df.iat[i, j])) is not None)]

    # Find the indices of columns containing the search value
    column_matches = [{'rowIndex': -1, 'colIndex': j} for j, column in enumerate(df.columns) if (re.search(search_regex,str(column)) is not None)]
    
    # We want the columns to come first
    all_matches = column_matches + cell_matches
    return {'total_number_matches': total_number_matches, 'matches': all_matches }



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.transpose_code_chunk import TransposeCodeChunk
from mitosheet.state import DATAFRAME_SOURCE_TRANSPOSED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.utils import get_first_unused_dataframe_name


class TransposeStepPerformer(StepPerformer):
    """
    Allows you to transpose.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'transpose'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:

        sheet_index: int = get_param(params, 'sheet_index')
        new_df_name = get_first_unused_dataframe_name(prev_state.df_names, f'{prev_state.df_names[sheet_index]}_transposed')
        execution_data = {
            'new_df_name': new_df_name,
        }

        return cls.execute_through_transpile(
            prev_state, 
            params, 
            execution_data, 
            {
                'df_source': DATAFRAME_SOURCE_TRANSPOSED,
                'new_df_names': [new_df_name],
                'overwrite': None
            }
        )

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            TransposeCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'), 
                get_param(execution_data if execution_data is not None else {}, 'new_df_name')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from mitosheet.step_performers.import_steps.excel_range_import import \
    ExcelRangeImportStepPerformer
from mitosheet.step_performers.import_steps.dataframe_import import \
    DataframeImportStepPerformer
from mitosheet.step_performers.import_steps.excel_import import \
    ExcelImportStepPerformer
from mitosheet.step_performers.import_steps.simple_import import \
    SimpleImportStepPerformer
from mitosheet.step_performers.import_steps.snowflake_import import SnowflakeImportStepPerformer
from mitosheet.step_performers.user_defined_import import UserDefinedImportStepPerformer


def is_import_step_type(step_type: str) -> bool:
    return step_type == SimpleImportStepPerformer.step_type() \
        or step_type == ExcelImportStepPerformer.step_type() \
        or step_type == DataframeImportStepPerformer.step_type() \
        or step_type == ExcelRangeImportStepPerformer.step_type() \
        or step_type == SnowflakeImportStepPerformer.step_type() \
        or step_type == UserDefinedImportStepPerformer.step_type()


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Replays an existing analysis
"""
from copy import copy, deepcopy
from typing import Any, Dict, List
from mitosheet.errors import make_no_analysis_error

from mitosheet.telemetry.telemetry_utils import log
from mitosheet.saved_analyses import read_and_upgrade_analysis
from mitosheet.types import StepsManagerType
from mitosheet.step_performers.import_steps import is_import_step_type
from mitosheet.api.get_imported_files_and_dataframes_from_current_steps import get_import_data_with_single_import_list
from mitosheet.updates.args_update import do_arg_update

REPLAY_ANALYSIS_UPDATE_EVENT = 'replay_analysis_update'
REPLAY_ANALYSIS_UPDATE_PARAMS = [
    'analysis_name',
    'step_import_data_list_to_overwrite',
    'args'
]

def overwrite_import_data(analysis: Dict[str, Any], step_import_data_list_to_overwrite: List[Dict[str, Any]]) -> Dict[str, Any]:
    """
    If there is import data to overwrite, we go through and overwrite it. We notably have to do this
    by keeping track of the index of the import, as we do not have step ids to link these things. 

    As this import data is split so that only one dataframe is created during any specific import,
    this means we make a bit more steps than we might want. But it's ok for now!
    """
    
    # Flatten all the imports into a single list, so that we can index into them. We notably
    # don't need the step ids for these
    all_imports = [_import for import_data in step_import_data_list_to_overwrite for _import in import_data['imports']]

    import_number = 0
    final_analysis: Dict[str, Any] = copy(analysis)
    final_analysis['steps_data'] = []

    # Rebuild all the steps in the analysis
    for step_data in analysis['steps_data']:
        step_type = step_data['step_type']
        params = step_data['params']

        if not is_import_step_type(step_type):
            # If it's not an import step, we don't need to change anything
            final_analysis['steps_data'].append(step_data)
        else:
            # If it is an import step, they we replace each dataframe creation with a single
            # step that is created with the new updated dataframe creation data 
            num_imports_in_step = len(get_import_data_with_single_import_list(step_type, params))
            for i in range(num_imports_in_step):
                import_data_index = import_number + i
                new_step_data = deepcopy(step_data)
                new_step_data['step_type'] = all_imports[import_data_index]['step_type']
                new_step_data['params'] = all_imports[import_data_index]['params']
                final_analysis['steps_data'].append(new_step_data)
            
            import_number += num_imports_in_step
            
    
    return final_analysis


def execute_replay_analysis_update(
        steps_manager: StepsManagerType,
        analysis_name: str,
        args: List[str],
        step_import_data_list_to_overwrite: List[Dict[str, Any]]
    ) -> None:
    """
    This function reapplies the analysis saved at analysis_name. Notably, if
    step_import_data_list_to_overwrite are passed, this it overwrites all the
    current import steps with the new updated import steps that are passed.

    This means that the number of dataframes created in step_import_data_list_to_overwrite,
    must be equal to the number of dataframes imported in analysis_name. The frontend
    should maintain this invariant when allowing users to update imports.

    If any step fails to execute, none of the analysis gets replayed at all.
    """
    # When the frontend is refreshed, we read in and replay the anlaysis again
    # unnecessarily, aka we need to read this
    if analysis_name == steps_manager.analysis_name:
        return

    # If we're getting an event telling us to update, we read in the steps from the file
    analysis = read_and_upgrade_analysis(analysis_name, args)

    # If there is no analysis with this name, generate an error
    if analysis is None:
        log('replayed_nonexistant_analysis_failed')
        raise make_no_analysis_error(analysis_name, error_modal=False)

    # If there is import data to overwrite, we do that
    if len(step_import_data_list_to_overwrite) > 0:
        analysis = overwrite_import_data(analysis, step_import_data_list_to_overwrite)

    previous_public_interface_version = steps_manager.public_interface_version
    previous_code_options = steps_manager.code_options

    # If there are args to the mitosheet call, then set them on the steps manager
    if len(args) > 0:
        do_arg_update(steps_manager, args)

    try:
        # Before we execute the steps, update to the public interface version of the saved analysis
        steps_manager.public_interface_version = analysis['public_interface_version']
        steps_manager.code_options = analysis['code_options']

        steps_manager.execute_steps_data(new_steps_data=analysis['steps_data'])

    except:
        # If we error, reset the public interface version, and code options
        steps_manager.public_interface_version = previous_public_interface_version
        steps_manager.code_options = previous_code_options
        raise

    # NOTE: We update the analysis name only if the new steps execute correctly,
    # so that we actually do go about overwriting the saved analysis in this case.
    # If the above errors, then we won't overwrite the analysis that was attempted 
    # to be played
    steps_manager.analysis_name = analysis_name


REPLAY_ANALYSIS_UPDATE = {
    'event_type': REPLAY_ANALYSIS_UPDATE_EVENT,
    'params': REPLAY_ANALYSIS_UPDATE_PARAMS,
    'execute': execute_replay_analysis_update
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from mitosheet.saved_analyses.save_utils import (
    read_and_upgrade_analysis, write_save_analysis_file,
    delete_saved_analysis, rename_saved_analysis, 
    _get_all_analysis_filenames, _delete_analyses,
    SAVED_ANALYSIS_FOLDER, read_analysis,
    get_steps_obj_for_saved_analysis, get_analysis_exists, 
    get_saved_analysis_string
)
from mitosheet.saved_analyses.upgrade import is_prev_version
from mitosheet.saved_analyses.upgrade import upgrade_saved_analysis_to_current_version




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import os
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.snowflake_import_code_chunk import SnowflakeImportCodeChunk
from mitosheet.errors import make_invalid_snowflake_credentials_error, make_invalid_snowflake_import_error

from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import ColumnHeader, SnowflakeCredentials, SnowflakeQueryParams, SnowflakeTableLocationAndWarehouse
from mitosheet.utils import get_valid_dataframe_name
from mitosheet.api.get_validate_snowflake_credentials import get_cached_snowflake_credentials

# The snowflake-connector-python package is only available in Python > 3.6 
# and is not distributed with the mitosheet package, so we make sure to 
# not import will succeed. 
try:
    import snowflake.connector
    SNOWFLAKE_CONNECTOR_IMPORTED = True
except ImportError:
    SNOWFLAKE_CONNECTOR_IMPORTED = False

class SnowflakeImportStepPerformer(StepPerformer):
    """
    Allows you to snowflakeimport.
    """

    @classmethod
    def step_version(cls) -> int:
        return 3

    @classmethod
    def step_type(cls) -> str:
        return 'snowflake_import'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:

        credentials = get_cached_snowflake_credentials()
        table_loc_and_warehouse: SnowflakeTableLocationAndWarehouse = get_param(params, 'table_loc_and_warehouse')
        query_params: SnowflakeQueryParams = get_param(params, 'query_params')

        if credentials is None:
            raise make_invalid_snowflake_credentials_error()    
            
        table_or_view = table_loc_and_warehouse['table_or_view']
        connection_params_dict = get_connection_param_dict(credentials, table_loc_and_warehouse)
        
        try: 
            # First try to establish the connection
            con = snowflake.connector.connect(**connection_params_dict)
        except Exception as e: 
            print(e)
            # When we do the frontend, we can figure out exactly what we want to raise here
            raise make_invalid_snowflake_import_error(e)
        
        sql_query = create_query(table_or_view, query_params)
        new_df_name = get_valid_dataframe_name(prev_state.df_names, table_or_view.lower())

        credentials_are_in_env = 'SNOWFLAKE_USERNAME' in os.environ and 'SNOWFLAKE_PASSWORD' in os.environ and 'SNOWFLAKE_ACCOUNT' in os.environ

        execution_data = {
            'connection_params_dict': connection_params_dict,
            'credentials_are_in_env': credentials_are_in_env,
            'sql_query': sql_query,
            'new_df_name': new_df_name
        }

        try:
            return cls.execute_through_transpile(
                prev_state,
                params,
                execution_data,
                new_dataframe_params={
                    'df_source': DATAFRAME_SOURCE_IMPORTED,
                    'new_df_names': [new_df_name],
                    'overwrite': None
                }
            )
        except Exception as e:
            raise make_invalid_snowflake_import_error(e)

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            SnowflakeImportCodeChunk(
                prev_state, 
                get_param(execution_data if execution_data is not None else {}, 'connection_params_dict'),
                get_param(execution_data if execution_data is not None else {}, 'credentials_are_in_env'),
                [get_param(execution_data if execution_data is not None else {}, 'sql_query')],
                [get_param(execution_data if execution_data is not None else {}, 'new_df_name')],
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}
    

def get_connection_param_dict (credentials: SnowflakeCredentials, table_loc_and_warehouse: SnowflakeTableLocationAndWarehouse) -> Dict[str, str]:
    return {
        'user': credentials['username'],
        'password': credentials['password'],
        'account': credentials['account'],
        'role': table_loc_and_warehouse['role'],
        'warehouse': table_loc_and_warehouse['warehouse'],
        'database': table_loc_and_warehouse['database'],
        'schema': table_loc_and_warehouse['schema'],
    }

def create_query(table_or_view: Optional[str], query_params: SnowflakeQueryParams) -> str:
    transpiled_column_headers = [get_snowflake_column_header(ch) for ch in query_params["columns"]]
    # TODO: When we add more options, let's make a helper function for building these strings
    limit = query_params.get('limit')
    limit_string = f' limit {limit}' if limit is not None else ''
    return f'SELECT {", ".join(transpiled_column_headers)} FROM {table_or_view}{limit_string}'

def get_snowflake_column_header(column_header: ColumnHeader) -> str:
    return f'\"{column_header}\"'



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.replace_code_chunk import ReplaceCodeChunk
from mitosheet.errors import make_invalid_replace_error
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param


class ReplaceStepPerformer(StepPerformer):
    """
    Allows you to replace a search value with a replace value in a sheet for both
    the values in the cells and the column headers.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'replace'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        try:
            post_state, execution_data = cls.execute_through_transpile(
                prev_state, 
                params
            )
        except Exception as e:
            search_value: str = get_param(params, 'search_value')
            replace_value: str = get_param(params, 'replace_value')
            raise make_invalid_replace_error(search_value, replace_value)
        
        return post_state, execution_data

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ReplaceCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_ids'),
                get_param(params, 'search_value'),
                get_param(params, 'replace_value'),
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}
    


import os
import string
from typing import Any, Dict, Union

from mitosheet.scheduling import get_automation_files_for_new_automation, create_github_pr

from mitosheet.types import StepsManagerType
from mitosheet.utils import get_valid_python_identifier


def get_pr_url_of_new_pr(params: Dict[str, Any], steps_manager: StepsManagerType) -> Union[str, Dict[str, str]]:
    automation_name = params['automation_name']
    automation_description = params['automation_description']
    schedule = params['schedule']

    try:
        automation_name_safe = get_valid_python_identifier(automation_name, 'automation', 'automation_')

        base_folder = f'automations/{automation_name_safe}'
        new_files = get_automation_files_for_new_automation(
            steps_manager, 
            automation_name,
            automation_description,
            base_folder,
            schedule
        )

        repo = os.environ.get('MITO_CONFIG_GITHUB_AUTOMATION_REPO')
        if not repo:
            raise Exception("GitHub repo not found in environment variables. See docs here: TODO: insert link")

        pr_url = create_github_pr(
            repo, 
            f'new-automation-{automation_name_safe}', 
            f'Add data, Python, and workflow files for automation: {automation_name}',
            f'Add automation: {automation_name}', 
            automation_description, 
            new_files
        )
    except Exception as e:
        return {
            'error': str(e)
        }

    return pr_url


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains all functions that can be used in a sheet that operate on dates.

NOTE: This file is alphabetical order!
"""
from datetime import datetime
from distutils.version import LooseVersion

import pandas as pd

from mitosheet.public.v3.errors import handle_sheet_function_errors
from mitosheet.public.v3.types.decorators import cast_values_in_arg_to_type
from mitosheet.public.v3.types.sheet_function_types import (
    DatetimeFunctionReturnType, DatetimeRestrictedInputType,
    IntFunctionReturnType, StringFunctionReturnType)


# Inspired by: https://stackoverflow.com/questions/69345845/why-does-dateoffset-rollback-not-work-the-way-i-expect-it-to-with-days-hours
# Given a datetime and a frequency (ie: business month begin), sets the 
# the timestamp back to the frequency unless its already there!
def to_start(t: pd.Timestamp, freq: pd.DateOffset) -> pd.Timestamp:
    try:
        return t.floor(freq) # fixed frequencies should just floor the date/time
    except ValueError: # if freq is variable, we fall into here...
        return freq.rollback(t.floor("D"))

def to_end(t: pd.Timestamp, freq: pd.DateOffset) -> pd.Timestamp:
    try:
        return t.ceil(freq) # fixed frequencies should just ceil the date/time
    except ValueError: # if freq is variable, we fall into here...
        return freq.rollforward(t.ceil("D"))


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def DATEVALUE(arg: DatetimeRestrictedInputType) -> DatetimeFunctionReturnType:
    """
    {
        "function": "DATEVALUE",
        "description": "Converts a given string to a date series.",
        "search_terms": ["datevalue", "date value", "date", "string to date", "datetime", "dtype", "convert"],
        "category": "DATE",
        "examples": [
            "DATEVALUE(date_column)",
            "DATEVALUE('2012-12-22')"
        ],
        "syntax": "DATEVALUE(date_string)",
        "syntax_elements": [{
                "element": "date_string",
                "description": "The date string to turn into a date object."
            }
        ]
    }
    """
    return arg


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def DAY(arg: DatetimeRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "DAY",
        "description": "Returns the day of the month that a specific date falls on, as a number.",
        "search_terms": ["day", "date"],
        "category": "DATE",
        "examples": [
            "DAY(date_column)",
            "DAY('2012-12-22')"
        ],
        "syntax": "DAY(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the day of."
            }
        ]
    }
    """
    if isinstance(arg, datetime) or isinstance(arg, pd.Timestamp):
        return arg.day
    
    return arg.dt.day
    

@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def ENDOFBUSINESSMONTH(arg: DatetimeRestrictedInputType) -> DatetimeFunctionReturnType:
    """
    {
        "function": "ENDOFBUSINESSMONTH",
        "description": "Given a date, returns the end of the buisness month. E.g. the last weekday.",
        "search_terms": ["business", "month", "eom", "eobm", "date", "workday", "end"],
        "category": "DATE",
        "examples": [
            "ENDOFBUSINESSMONTH(date_column)",
            "ENDOFBUSINESSMONTH('2012-12-22')"
        ],
        "syntax": "ENDOFBUSINESSMONTH(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the end of the business month of."
            }
        ]
    }
    """
    if isinstance(arg, pd.Timestamp):
        return to_end(arg, pd.tseries.offsets.BMonthEnd(n=0))
    if isinstance(arg, datetime):
        return to_end(pd.Timestamp(arg), pd.tseries.offsets.BMonthEnd(n=0))
    
    return arg.apply(lambda t: to_end(t, pd.tseries.offsets.BMonthEnd(n=0)))



@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def ENDOFMONTH(arg: DatetimeRestrictedInputType) -> DatetimeFunctionReturnType:
    """
    {
        "function": "ENDOFMONTH",
        "description": "Given a date, returns the end of the month, as a date. E.g. input of 12-22-1997 will return 12-31-1997.",
        "search_terms": ["month", "eom", "date", "workday", "end", "eomonth"],
        "category": "DATE",
        "examples": [
            "ENDOFMONTH(date_column)",
            "ENDOFMONTH('2012-12-22')"
        ],
        "syntax": "ENDOFMONTH(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the last day of the month of."
            }
        ]
    }
    """
    if isinstance(arg, pd.Timestamp):
        return to_end(arg, pd.tseries.offsets.MonthEnd(n=0))
    if isinstance(arg, datetime):
        return to_end(pd.Timestamp(arg), pd.tseries.offsets.MonthEnd(n=0))
    
    return arg.apply(lambda t: to_end(t, pd.tseries.offsets.MonthEnd(n=0)))


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def HOUR(arg: DatetimeRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "HOUR",
        "description": "Returns the hour component of a specific date, as a number.",
        "search_terms": ["hour", "hr"],
        "category": "DATE",
        "examples": [
            "HOUR(date_column)",
            "HOUR('2012-12-22 09:45:00')"
        ],
        "syntax": "HOUR(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the hour of."
            }
        ]
    }
    """
    if isinstance(arg, datetime) or isinstance(arg, pd.Timestamp):
        return arg.hour
    
    return arg.dt.hour


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def MINUTE(arg: DatetimeRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "MINUTE",
        "description": "Returns the minute component of a specific date, as a number.",
        "search_terms": ["minute", "min"],
        "category": "DATE",
        "examples": [
            "MINUTE(date_column)",
            "MINUTE('2012-12-22 09:45:00')"
        ],
        "syntax": "MINUTE(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the minute of."
            }
        ]
    }
    """
    if isinstance(arg, datetime) or isinstance(arg, pd.Timestamp):
        return arg.minute
    
    return arg.dt.minute


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def MONTH(arg: DatetimeRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "MONTH",
        "description": "Returns the month that a specific date falls in, as a number.",
        "search_terms": ["month", "date"],
        "category": "DATE",
        "examples": [
            "MONTH(date_column)",
            "MONTH('2012-12-22')"
        ],
        "syntax": "MONTH(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the month of."
            }
        ]
    }
    """
    if isinstance(arg, datetime) or isinstance(arg, pd.Timestamp):
        return arg.month
    
    return arg.dt.month


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def MONTHNAME(arg: DatetimeRestrictedInputType) -> StringFunctionReturnType:
    """
    {
        "function": "MONTHNAME",
        "description": "Returns the month that a specific date falls in, as Jan, Feb, Mar, etc.",
        "search_terms": ["month", "monthname", "date"],
        "category": "DATE",
        "examples": [
            "MONTHNAME(date_column)",
            "MONTHNAME('2012-12-22')"
        ],
        "syntax": "MONTHNAME(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the month of."
            }
        ]
    }
    """
    if isinstance(arg, datetime) or isinstance(arg, pd.Timestamp):
        return arg.strftime('%b')

    return arg.dt.strftime('%b')


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def QUARTER(arg: DatetimeRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "QUARTER",
        "description": "Returns the quarter (1-4) that a specific date falls in, as a number.",
        "search_terms": ["quarter"],
        "category": "DATE",
        "examples": [
            "QUARTER(date_column)",
            "QUARTER('2012-12-22')"
        ],
        "syntax": "QUARTER(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the quarter of."
            }
        ]
    }
    """
    if isinstance(arg, datetime):
        return pd.Timestamp(arg).quarter
    elif isinstance(arg, pd.Timestamp):
        return arg.quarter
    
    return arg.dt.quarter


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def STARTOFBUSINESSMONTH(arg: DatetimeRestrictedInputType) -> DatetimeFunctionReturnType:
    """
    {
        "function": "STARTOFBUSINESSMONTH",
        "description": "Given a date, returns the most recent start of the business month, as a state. E.g. the first weekday.",
        "search_terms": ["business", "month", "SOM", "SOBM", "date", "start"],
        "category": "DATE",
        "examples": [
            "STARTOFBUSINESSMONTH(date_column)",
            "STARTOFBUSINESSMONTH('2012-12-22 09:23:05')"
        ],
        "syntax": "STARTOFBUSINESSMONTH(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the most recent beginning of month business day of."
            }
        ]
    }
    """
    if isinstance(arg, pd.Timestamp):
        return to_start(arg, pd.tseries.offsets.BMonthBegin(n=1))
    elif isinstance(arg, datetime):
        return to_start(pd.Timestamp(arg), pd.tseries.offsets.BMonthBegin(n=1))
    
    return arg.apply(lambda t: to_start(t, pd.tseries.offsets.BMonthBegin(n=1)))
    

@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def STARTOFMONTH(arg: DatetimeRestrictedInputType) -> DatetimeFunctionReturnType:
    """
    {
        "function": "STARTOFMONTH",
        "description": "Given a date, returns the start of the month, as a date. E.g. input of 12-22-1997 will return 12-1-1997.",
        "search_terms": ["month", "SOM", "date", "start"],
        "category": "DATE",
        "examples": [
            "STARTOFMONTH(date_column)",
            "STARTOFMONTH('2012-12-22 09:23:05')"
        ],
        "syntax": "STARTOFMONTH(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the first day of the month of."
            }
        ]
    }
    """
    if isinstance(arg, pd.Timestamp):
        return to_start(arg, pd.tseries.offsets.MonthBegin(n=1))
    elif isinstance(arg, datetime):
        return to_start(pd.Timestamp(arg), pd.tseries.offsets.MonthBegin(n=1))
    
    return arg.apply(lambda t: to_start(t, pd.tseries.offsets.MonthBegin(n=1)))



@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def STRIPTIMETOMINUTES(arg: DatetimeRestrictedInputType) -> DatetimeFunctionReturnType:
    """
    {
        "function": "STRIPTIMETOMINUTES",
        "description": "Returns the date with a seconds component of 00.",
        "search_terms": ["time", "date", "minutes", "strip"],
        "category": "DATE",
        "examples": [
            "STRIPTIMETOMINUTES(date_column)",
            "STRIPTIMETOMINUTES('2012-12-22 09:23:05')"
        ],
        "syntax": "STRIPTIMETOMINUTES(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to reset the seconds component of."
            }
        ]
    }
    """
    if isinstance(arg, pd.Timestamp):
        return arg.floor('Min')
    elif isinstance(arg, datetime):
        return pd.Timestamp(arg).floor('Min')
    
    return arg.dt.floor('Min')


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def STRIPTIMETOHOURS(arg: DatetimeRestrictedInputType) -> DatetimeFunctionReturnType:
    """
    {
        "function": "STRIPTIMETOHOURS",
        "description": "Returns the date with a seconds and minutes component of 00:00.",
        "search_terms": ["time", "date", "hours", "strip"],
        "category": "DATE",
        "examples": [
            "STRIPTIMETOHOURS(date_column)",
            "STRIPTIMETOHOURS('2012-12-22 09:23:05')"
        ],
        "syntax": "STRIPTIMETOHOURS(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to reset the seconds and minutes component of."
            }
        ]
    }
    """
    if isinstance(arg, pd.Timestamp):
        return arg.floor('H')
    elif isinstance(arg, datetime):
        return pd.Timestamp(arg).floor('H')
    
    return arg.dt.floor('H')


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def STRIPTIMETODAYS(arg: DatetimeRestrictedInputType) -> DatetimeFunctionReturnType:
    """
    {
        "function": "STRIPTIMETODAYS",
        "description": "Returns the date with a seconds, minutes, and hours component of 00:00:00.",
        "search_terms": ["time", "date", "days", "strip"],
        "category": "DATE",
        "examples": [
            "STRIPTIMETODAYS(date_column)",
            "STRIPTIMETODAYS('2012-12-22 09:23:05')"
        ],
        "syntax": "STRIPTIMETODAYS(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to reset the seconds, minutes, and hours component of."
            }
        ]
    }
    """
    if isinstance(arg, pd.Timestamp):
        return arg.floor('D')
    elif isinstance(arg, datetime):
        return pd.Timestamp(arg).floor('D')
    
    return arg.dt.floor('D')


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def STRIPTIMETOMONTHS(arg: DatetimeRestrictedInputType) -> DatetimeFunctionReturnType:
    """
    {
        "function": "STRIPTIMETOMONTHS",
        "description": "Returns the date adjusted to the start of the month.",
        "search_terms": ["time", "date", "months", "strip"],
        "category": "DATE",
        "examples": [
            "STRIPTIMETOMONTHS(date_column)",
            "STRIPTIMETOMONTHS('2012-12-22 09:23:05')"
        ],
        "syntax": "STRIPTIMETOMONTHS(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to reset the seconds, minutes, hours, and days of."
            }
        ]
    }
    """
    if isinstance(arg, pd.Timestamp):
        return arg.floor('D') - pd.tseries.offsets.MonthBegin()
    elif isinstance(arg, datetime):
        return pd.Timestamp(arg).floor('D') - pd.tseries.offsets.MonthBegin()
    
    return arg.dt.floor('D') - pd.tseries.offsets.MonthBegin()

@handle_sheet_function_errors
def TODAY() -> DatetimeFunctionReturnType:
    """
    {
        "function": "TODAY",
        "description": "Returns the datetime value for the current date.",
        "search_terms": ["today"],
        "category": "DATE",
        "examples": [
            "TODAY()"
        ],
        "syntax": "TODAY()",
        "syntax_elements": []
    }
    """
    return pd.to_datetime('today').normalize()

@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def STRIPTIMETOYEARS(arg: DatetimeRestrictedInputType) -> DatetimeFunctionReturnType:
    """
    {
        "function": "STRIPTIMETOYEARS",
        "description": "Returns the date adjusted to the start of the year.",
        "search_terms": ["time", "date", "years", "strip"],
        "category": "DATE",
        "examples": [
            "STRIPTIMETOYEARS(date_column)",
            "STRIPTIMETOYEARS('2012-12-22 09:23:05')"
        ],
        "syntax": "STRIPTIMETOYEARS(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to reset the seconds, minutes, hours, days, and month components of."
            }
        ]
    }
    """
    if isinstance(arg, pd.Timestamp):
        return arg.floor('D') - pd.tseries.offsets.YearBegin()
    elif isinstance(arg, datetime):
        return pd.Timestamp(arg).floor('D') - pd.tseries.offsets.YearBegin()
    
    return arg.dt.floor('D') - pd.tseries.offsets.YearBegin()


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def SECOND(arg: DatetimeRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "SECOND",
        "description": "Returns the seconds component of a specific date, as a number.",
        "search_terms": ["second", "sec"],
        "category": "DATE",
        "examples": [
            "SECOND(date_column)",
            "SECOND('2012-12-22 09:23:05')"
        ],
        "syntax": "SECOND(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the seconds of."
            }
        ]
    }
    """
    if isinstance(arg, datetime):
        return pd.Timestamp(arg).second
    elif isinstance(arg, pd.Timestamp):
        return arg.second

    return arg.dt.second


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def WEEK(arg: DatetimeRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "WEEK",
        "description": "Returns the week (1-52) of a specific date, as a number.",
        "search_terms": ["week", "1", "52"],
        "category": "DATE",
        "examples": [
            "WEEK(date_column)",
            "WEEK('2012-12-22 09:23:05')"
        ],
        "syntax": "WEEK(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the week of."
            }
        ]
    }
    """
    if isinstance(arg, datetime):
        return pd.Timestamp(arg).week
    elif isinstance(arg, pd.Timestamp):
        return arg.week

    # Handle if we're on pandas version < 1.1, where isocalendar() is not available
    if LooseVersion(pd.__version__) < LooseVersion('1.1'):
        return arg.dt.week
    
    return arg.dt.isocalendar().week


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def WEEKDAY(arg: DatetimeRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "WEEKDAY",
        "description": "Returns the day of the week that a specific date falls on. 1-7 corresponds to Monday-Sunday.",
        "search_terms": ["weekday", "monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"],
        "category": "DATE",
        "examples": [
            "WEEKDAY(date_column)",
            "WEEKDAY('2012-12-22')"
        ],
        "syntax": "WEEKDAY(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the weekday of."
            }
        ]
    }
    """
    if isinstance(arg, datetime):
        return pd.Timestamp(arg).weekday() + 1
    elif isinstance(arg, pd.Timestamp):
        return arg.weekday() + 1

    return arg.dt.weekday + 1


@cast_values_in_arg_to_type('arg', 'datetime')
@handle_sheet_function_errors
def YEAR(arg: DatetimeRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "YEAR",
        "description": "Returns the day of the year that a specific date falls in, as a number.",
        "search_terms": ["year", "date"],
        "category": "DATE",
        "examples": [
            "YEAR(date_column)",
            "YEAR('2012-12-22')"
        ],
        "syntax": "YEAR(date)",
        "syntax_elements": [{
                "element": "date",
                "description": "The date or date series to get the month of."
            }
        ]
    }
    """
    if isinstance(arg, datetime):
        return pd.Timestamp(arg).year
    elif isinstance(arg, pd.Timestamp):
        return arg.year

    return arg.dt.year


DATE_FUNCTIONS = {
    'DATEVALUE': DATEVALUE,
    'DAY': DAY,
    'ENDOFBUSINESSMONTH': ENDOFBUSINESSMONTH,
    'ENDOFMONTH': ENDOFMONTH,
    'HOUR': HOUR,
    'MONTH': MONTH,
    'MONTHNAME': MONTHNAME,
    'MINUTE': MINUTE,
    'QUARTER': QUARTER,
    'SECOND': SECOND,
    'STARTOFBUSINESSMONTH': STARTOFBUSINESSMONTH, 
    'STARTOFMONTH': STARTOFMONTH, 
    'STRIPTIMETOMINUTES': STRIPTIMETOMINUTES, 
    'STRIPTIMETOHOURS': STRIPTIMETOHOURS,
    'STRIPTIMETODAYS': STRIPTIMETODAYS,
    'STRIPTIMETOMONTHS': STRIPTIMETOMONTHS,
    'STRIPTIMETOYEARS': STRIPTIMETOYEARS,
    'TODAY': TODAY,
    'WEEK': WEEK,
    'WEEKDAY': WEEKDAY,
    'YEAR': YEAR,
}


import json
from typing import Any, Dict, List, Optional
import pandas as pd
from mitosheet.types import ConditionalFormattingCellResults, ConditionalFormattingInvalidResults, ConditionalFormattingResult, StateType
from mitosheet.utils import MAX_ROWS, NpEncoder


def get_conditonal_formatting_result(
        state: StateType,
        sheet_index: int,
        df: pd.DataFrame,
        conditional_formatting_rules: List[Dict[str, Any]],
        max_rows: Optional[int]=MAX_ROWS,
    ) -> ConditionalFormattingResult: 
    from mitosheet.step_performers.filter import check_filters_contain_condition_that_needs_full_df

    invalid_conditional_formats: ConditionalFormattingInvalidResults = dict()
    formatted_result: ConditionalFormattingCellResults = dict()

    for conditional_format in conditional_formatting_rules:
        try:

            format_uuid = conditional_format["format_uuid"]
            column_ids = conditional_format["columnIDs"]

            for column_id in column_ids:
                if column_id not in formatted_result:
                    formatted_result[column_id] = dict()

                filters  = conditional_format["filters"]
                backgroundColor = conditional_format.get("backgroundColor", None)
                color = conditional_format.get("color", None)
                
                # Certain filter conditions require the entire dataframe to be present, as they calculate based
                # on the full dataframe. In other cases, we only operate on the first 1500 rows, for speed
                _df = df
                if not check_filters_contain_condition_that_needs_full_df(filters):
                    df = df.head(max_rows)

                column_header = state.column_ids.get_column_header_by_id(sheet_index, column_id)

                # Use the get_applied_filter function from our filtering infrastructure
                from mitosheet.step_performers.filter import \
                    get_full_applied_filter
                full_applied_filter, _ = get_full_applied_filter(_df, column_header, 'And', filters)

                # We can only take the first max_rows here, as this is all we need
                applied_indexes = _df[full_applied_filter].head(max_rows).index.tolist()

                for index in applied_indexes:
                    # We need to make this index valid json, and do so in a way that is consistent with how indexes
                    # are sent to the frontend. However, if the index is a string, we don't want to add quotes around it
                    if isinstance(index, str):
                        json_index = index
                    else:
                        json_index = json.dumps(index, cls=NpEncoder)
                    formatted_result[column_id][json_index] = {'backgroundColor': backgroundColor, 'color': color}

        except Exception as e:
            if format_uuid not in invalid_conditional_formats:
                invalid_conditional_formats[format_uuid] = []
            invalid_conditional_formats[format_uuid].append(column_id)

    return {
        'invalid_conditional_formats': invalid_conditional_formats,
        'results': formatted_result
    }


import json
from mitosheet.mito_backend import MitoBackend
from mitosheet.saved_analyses import get_saved_analysis_string


def flatten_mito_backend_to_json(mito_backend: MitoBackend) -> str:
    saved_analysis_string = get_saved_analysis_string(mito_backend.steps_manager)
    return json.dumps({
        'backend_state': saved_analysis_string,
        'shared_state_variables': mito_backend.get_shared_state_variables()
    })

def read_backend_state_string_to_mito_backend(backend_state_string: str) -> MitoBackend:

    mb = MitoBackend()
    steps_manager = mb.steps_manager
    analysis = json.loads(backend_state_string)

    previous_public_interface_version = steps_manager.public_interface_version
    previous_code_options = steps_manager.code_options

    try:
        
        # Before we execute the steps, update to the public interface version of the saved analysis
        steps_manager.public_interface_version = analysis['public_interface_version']
        steps_manager.code_options = analysis['code_options']

        steps_manager.execute_steps_data(new_steps_data=analysis['steps_data'])

    except:
        # If we error, reset the public interface version, and code options
        steps_manager.public_interface_version = previous_public_interface_version
        steps_manager.code_options = previous_code_options
        raise  

    return mb



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
import pandas as pd
from typing import Any, Dict, List, Optional
import plotly.express as px
import plotly.graph_objects as go
from mitosheet.step_performers.graph_steps.graph_utils import (
    get_html_and_script_from_figure,
)
from mitosheet.is_type_utils import is_number_dtype
from mitosheet.types import ColumnHeader, ColumnID
from mitosheet.telemetry.telemetry_utils import log
from mitosheet.steps_manager import StepsManager
from mitosheet.step_performers.bulk_old_rename.deprecated_utils import deprecated
from mitosheet.step_performers.graph_steps.graph_utils import BAR, BOX, SCATTER

# Max number of unique non-number items to display in a graph.
# NOTE: make sure to change both in unison so they make sense
MAX_UNIQUE_NON_NUMBER_VALUES = 10_000


def get_column_summary_graph(params: Dict[str, Any], steps_manager: StepsManager) -> Dict[str, Any]:
    """
    Creates a column summary graph and sends it back as a PNG
    string to the frontend for display.
    """
    sheet_index = params['sheet_index']
    column_id: ColumnID = params['column_id']
    height = params['height']
    width = params['width']
    include_plotlyjs = params['include_plotlyjs']


    # Create a copy of the dataframe, just for safety.
    df: pd.DataFrame = steps_manager.dfs[sheet_index].copy()

    column_header = steps_manager.curr_step.final_defined_state.column_ids.get_column_header_by_id(sheet_index, column_id)
    fig = _get_column_summary_graph(df, column_header)
        
    # Get rid of some of the default white space
    fig.update_layout(
        margin=dict(
            l=0,
            r=0,
            t=30,
            b=30,
        )
    )

    return_object = get_html_and_script_from_figure(fig, height, width, include_plotlyjs)

    return return_object

def filter_df_to_top_unique_values_in_series(
    df: pd.DataFrame,
    main_series: pd.Series,
    num_unique_values: int,
) -> pd.Series:
    """
    Helper function for filtering the dataframe down to the top most common
    num_unique_values in the main_series. Will not change the series if there are less
    values than that.

    The function filters the entire dataframe to make sure that the columns stay
    the same length (which is necessary if you want to graph them).

    It returns the filtered dataframe
    """
    if (
        len(main_series) < num_unique_values
        or main_series.nunique() < num_unique_values
    ):
        return df

    value_counts_series = main_series.value_counts()
    most_frequent_values_list = value_counts_series.head(
        n=num_unique_values
    ).index.tolist()

    return df[main_series.isin(most_frequent_values_list)]


def _get_column_summary_graph(df: pd.DataFrame, column_header: ColumnHeader) -> go.Figure:
    """
    One Axis Graphs heuristics:
    1. Number Column - we do no filtering. These graphs are pretty efficient up to 1M rows
    2. Non-number column. We filter to the top 10k values, as the graphs get pretty laggy
       beyond that
    """
    series: pd.Series = df[column_header]
    column_dtype = str(series.dtype)

    graph_title = f"{column_header} Frequencies"

    filtered = False
    if not is_number_dtype(column_dtype):
        if series.nunique() > MAX_UNIQUE_NON_NUMBER_VALUES:
            df = filter_df_to_top_unique_values_in_series(
                df, series, MAX_UNIQUE_NON_NUMBER_VALUES
            )
            # Set series as the newly filtered series
            series = df[column_header]

            filtered = True

        # Fill NaN values with 'NaN' so they are displayed in the graph.
        series = series.fillna("NaN")

    labels = {"x": ""}

    kwargs = {
        "x": series,
        "labels": labels,
        "title": graph_title,
    }

    fig = px.histogram(**kwargs)

    log(f"generate_column_summary_stat_graph", {"param_filtered": filtered})

    return fig



"""
PRESERVED FOR BACKWARDS COMPATABILITY. 
This function is exported to users in the generated code. We will preserve it here so that 
the generated code of our users still executes without error. 

We also preserve the constants that the function utilizes.
"""

SCATTER_PLOT_LESS_THAN_4_SERIES_MAX_NUMBER_OF_ROWS = 10_000
SCATTER_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS = 5_000
BAR_CHART_MAX_NUMBER_OF_ROWS = 5_000
BOX_PLOT_3_SERIES_MAX_NUMBER_OF_ROWS = 500_000
BOX_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS = 250_000


def filter_df_to_safe_size(
    graph_type: str,
    df: pd.DataFrame,
    column_headers: List[ColumnHeader],
    other_axis_column_headers: Optional[List[ColumnHeader]] = None,
) -> pd.DataFrame:
    """
    A helper function that filters a dataframe down to a safe size
    to display in a graph, depending on the type of graph.

    This is used to stop the graph from crashing the users browser
    when it displays too much data.

    It is also exported from the mitosheet package so that it can
    be used in the code that is exported from graphs.
    """
    original_df_len = len(df)

    if graph_type == BOX:
        if len(column_headers) <= 2:
            return df, False
        elif len(column_headers) == 3:
            return (
                df.head(BOX_PLOT_3_SERIES_MAX_NUMBER_OF_ROWS),
                original_df_len > BOX_PLOT_3_SERIES_MAX_NUMBER_OF_ROWS,
            )
        else:
            return (
                df.head(BOX_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS),
                original_df_len > BOX_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS,
            )
    elif graph_type == SCATTER:
        for column_header in column_headers:
            if not is_number_dtype(str(df[column_header].dtype)):
                # For each non-number series, filter it to only contain the most common values
                df = filter_df_to_top_unique_values_in_series(
                    df,
                    df[column_header],
                    MAX_UNIQUE_NON_NUMBER_VALUES,
                )

        # Then, take the first 10k rows
        total_allowed_rows = (
            SCATTER_PLOT_LESS_THAN_4_SERIES_MAX_NUMBER_OF_ROWS
            if len(column_headers) < 4
            else SCATTER_PLOT_4_SERIES_MAX_NUMBER_OF_ROWS
        )
        df = df.head(total_allowed_rows)

        return df, len(df) < original_df_len
    elif graph_type == BAR:
        # If both axises are supplied, then we go on to create the bar chart.
        # Start by sorting the key column in decreasing frequency so we can filter
        # dataframe according to the top BAR_CHART_MAX_NUMBER_OF_ROWS

        # TODO: this is bugged, with multi-column headers. For now, we just take
        # the max number of rows

        # Take the first BAR_CHART_MAX_NUMBER_OF_ROWS rows
        df = df.head(BAR_CHART_MAX_NUMBER_OF_ROWS)

        return df, len(df) < original_df_len


filter_df_to_safe_size_external = deprecated(filter_df_to_safe_size)


from dash import Dash, html, dcc, Output, Input
from dash.exceptions import PreventUpdate
import plotly.express as px
import pandas as pd

from mitosheet.mito_dash.v1 import Spreadsheet, mito_callback, activate_mito

df = pd.read_csv('https://raw.githubusercontent.com/plotly/datasets/master/gapminder_unfiltered.csv')

app = Dash(__name__)
activate_mito(app)

app.layout = html.Div([
    Spreadsheet(df, id={'type': 'spreadsheet', 'id': 'sheet'}, track_selection=True),
    dcc.Graph(id='graph-content'),
])

@mito_callback(
    Output('graph-content', 'figure'),
    Input({'type': 'spreadsheet', 'id': 'sheet'}, 'spreadsheet_selection')
)
def update_code(spreadsheet_selection):
    if spreadsheet_selection is None:
        raise PreventUpdate

    # Get selected rows from the original dataframe
    index = spreadsheet_selection.index
    dff = df.loc[index]

    return px.line(dff, x='year', y='lifeExp', color='continent')

if __name__ == '__main__':
    app.run(debug=True)



#!/usr/bin/env python
# coding: utf-8

from copy import copy
# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import get_param_dict_as_code


class SnowflakeImportCodeChunk(CodeChunk):

    def __init__(
        self, 
        prev_state: State, 
        connection_params_dict: Dict[str, str], 
        credentials_are_in_env: bool,
        sql_queries: List[str],
        new_df_names: List[str]
    ):
        super().__init__(prev_state)
        self.connection_params_dict = connection_params_dict
        self.credentials_are_in_env = credentials_are_in_env
        self.sql_queries = sql_queries

        self.new_df_names = new_df_names


    def get_display_name(self) -> str:
        return 'Import from Snowflake'
    
    def get_description_comment(self) -> str:
        return "Imported dataframe from Snowflake"

    def get_code(self) -> Tuple[List[str], List[str]]:

        imports = ['import snowflake.connector']

        if not self.credentials_are_in_env:
            connection_param_transpiled_code = get_param_dict_as_code(self.connection_params_dict)
            snowflake_connection_code = [
                f'con = snowflake.connector.connect({connection_param_transpiled_code})',
                '',
                'cur = con.cursor()',
            ]
        else:
            # Get rid of user, password, and account
            new_connection_params = copy(self.connection_params_dict)
            new_connection_params.pop('user')
            new_connection_params.pop('password')
            new_connection_params.pop('account')

            connection_param_transpiled_code = get_param_dict_as_code(new_connection_params, tab_level=0)

            imports.append('import os')
            snowflake_connection_code = [
                'con = snowflake.connector.connect(',
                f'    user=os.environ[\'SNOWFLAKE_USERNAME\'],',
                f'    password=os.environ[\'SNOWFLAKE_PASSWORD\'],',
                f'    account=os.environ[\'SNOWFLAKE_ACCOUNT\'],',
                f'    {connection_param_transpiled_code.strip()}',
                ')',
                '',
                'cur = con.cursor()',
            ]

        df_creation_code = ['']
        for sql_query, df_name in zip(self.sql_queries, self.new_df_names):
            df_creation_code.append(f'cur.execute(\'{sql_query}\')')
            df_creation_code.append(f'{df_name} = cur.fetch_pandas_all()')
            df_creation_code.append('')

        close_connection_code = ['con.close()']

        all_code = snowflake_connection_code + df_creation_code + close_connection_code
                        
        return all_code, imports

    def get_created_sheet_indexes(self) -> List[int]:
        return [i for i in range(len(self.prev_state.dfs), len(self.prev_state.dfs) + len(self.new_df_names))]

    def _combine_right_with_snowflake_import_code_chunk(self, other_code_chunk: "SnowflakeImportCodeChunk") -> Optional["SnowflakeImportCodeChunk"]:
        if not self.params_match(other_code_chunk, ['connection_params_dict']):
            return None
        
        all_sql_queries = copy(self.sql_queries) # Make sure to copy this so we don't get weird bugs w/ duplication
        all_sql_queries.extend(other_code_chunk.sql_queries)

        all_df_names = copy(self.new_df_names)
        all_df_names.extend(other_code_chunk.new_df_names)
        
        return SnowflakeImportCodeChunk(
            self.prev_state,
            self.connection_params_dict,
            self.credentials_are_in_env,
            all_sql_queries,
            all_df_names
        )        

    def combine_right(self, other_code_chunk: CodeChunk) -> Optional[CodeChunk]:
        if isinstance(other_code_chunk, SnowflakeImportCodeChunk):
            return self._combine_right_with_snowflake_import_code_chunk(other_code_chunk)
            
        return None
        


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import deepcopy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.dataframe_steps.dataframe_duplicate_code_chunk import \
    DataframeDuplicateCodeChunk
from mitosheet.column_headers import get_column_header_id
from mitosheet.state import DATAFRAME_SOURCE_DUPLICATED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.utils import get_first_unused_dataframe_name


class DataframeDuplicateStepPerformer(StepPerformer):
    """
    This steps duplicates a dataframe of a given index. 
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'dataframe_duplicate'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')

        post_state = prev_state.copy()

        # Execute the step
        pandas_start_time = perf_counter()
        df_copy = post_state.dfs[sheet_index].copy(deep=True)
        pandas_processing_time = perf_counter() - pandas_start_time
        new_name = get_first_unused_dataframe_name(post_state.df_names, post_state.df_names[sheet_index] + '_copy')

        # Copy the formatting to the new sheet. Because the mapping is column id -> format object, and
        # the column ids that are created for the df_copy in the add_df_to_state function might be different
        # than the column_ids created initially (e.g. because of renames), we have to go through and updated
        # the mapping with the new column ids that the format types must rely on
        old_df_format = post_state.df_formats[sheet_index]
        new_df_format = deepcopy(post_state.df_formats[sheet_index])
        for old_column_id, column_header in post_state.column_ids.get_column_ids_map(sheet_index).items():
            new_column_id = get_column_header_id(column_header)
            old_column_format = old_df_format['columns'].get(old_column_id, None)
            if old_column_format is not None:
                del new_df_format['columns'][old_column_id]
                new_df_format['columns'][new_column_id] = old_column_format

        post_state.add_df_to_state(df_copy, DATAFRAME_SOURCE_DUPLICATED, df_name=new_name, df_format=new_df_format)

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'new_df_name': new_name
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            DataframeDuplicateCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'),
                get_param(execution_data if execution_data is not None else {}, 'new_df_name')
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains tests for the user utils, which determine the current
state of the user (e.g. should they upgrade).
"""
import os
from datetime import datetime, timedelta

import pytest

from mitosheet.user import (is_local_deployment,
                            is_on_kuberentes_mito, is_running_test)


def test_is_local():
    assert is_local_deployment()
    assert not is_on_kuberentes_mito()

def test_detects_tests():
    assert is_running_test()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import traceback
from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.user_defined_import_code_chunk import \
    UserDefinedImportCodeChunk
from mitosheet.errors import MitoError
from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.user_defined_function_utils import \
    get_user_defined_function_param_type_and_execute_value_and_transpile_value
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.utils import get_first_unused_dataframe_name


class UserDefinedImportStepPerformer(StepPerformer):
    """
    Allows a user to use one of the user_defined_importers.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'user_defined_import'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        importer: str = get_param(params, 'importer')
        importer_params: Dict[str, str] = get_param(params, 'importer_params')
        
        new_df_name = get_first_unused_dataframe_name(prev_state.df_names, 'df' + str(len(prev_state.df_names)))

        importer_function = next(f for f in prev_state.user_defined_importers if f.__name__ == importer)

        execution_data = {
            'user_defined_function_params': get_user_defined_function_param_type_and_execute_value_and_transpile_value(prev_state, importer_function, importer_params),
            'new_df_names': [new_df_name],
        }
        
        try:
            return cls.execute_through_transpile(
                prev_state,
                params,
                execution_data,
                {
                    'df_source': DATAFRAME_SOURCE_IMPORTED,
                    'new_df_names': [new_df_name],
                    'overwrite': None
                }
            )
        except:
            traceback_final_line = traceback.format_exc().splitlines()[-1]

            raise MitoError(
                'user_defined_importer_error',
                f"Importer {importer} raised an error.",
                f"User defined importer {importer} raised an error: {traceback_final_line}",
                error_modal=False
            )

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            UserDefinedImportCodeChunk(
                prev_state, 
                get_param(params, 'importer'),
                get_param(execution_data if execution_data is not None else {}, 'user_defined_function_params'),
                get_param(execution_data if execution_data is not None else {}, 'new_df_names')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.empty_code_chunk import EmptyCodeChunk
from mitosheet.state import State
from mitosheet.step_performers.graph_steps.graph_utils import get_graph_index_by_graph_id
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import GraphID


class GraphDeleteStepPerformer(StepPerformer):
    """
    Deletes a graph from everywhere in the step.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'graph_delete'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        graph_id: GraphID = get_param(params, 'graph_id')

        post_state = prev_state.copy()

        graph_index = get_graph_index_by_graph_id(post_state.graph_data_array, graph_id)
        del post_state.graph_data_array[graph_index]

        return post_state, {
            'pandas_processing_time': 0 # No time spent on pandas, only metadata changes
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:

        # Graph steps don't add any generated code to the analysis script. 
        return [
            EmptyCodeChunk(
                prev_state, 
                'Deleted graph',
                'Deleted a graph',
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import re
from distutils.version import LooseVersion
from typing import List, Tuple, Any
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.types import ColumnID
from mitosheet.transpiler.transpile_utils import get_column_header_list_as_transpiled_code, get_column_header_map_as_code_string
from mitosheet.state import State

import pandas as pd

from mitosheet.errors import MitoError


def convert_to_original_type_or_str(column: str, original_type: type) -> Any:
    try:
        return original_type(column)
    except:
        return column


class ReplaceCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, sheet_index: int, column_ids: List[ColumnID], search_value: str, replace_value: str):
        super().__init__(prev_state)
        self.sheet_index = sheet_index
        self.column_ids = column_ids
        self.search_value = search_value
        self.replace_value = replace_value
        self.df = self.prev_state.dfs[self.sheet_index]
        self.df_name = self.prev_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'Replace'
    
    def get_description_comment(self) -> str:
        return f"Replace {self.search_value} with {self.replace_value} in {self.prev_state.df_names[self.sheet_index]}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        search_value = self.search_value
        replace_value = self.replace_value
        column_ids = self.column_ids
        df_name = self.df_name
        df_name_with_selected_columns = self.df_name
        sheet_index = self.sheet_index
        df = self.df
        column_headers: Any = []

        if (column_ids is not None and len(column_ids) > 0):
            column_headers = self.prev_state.column_ids.get_column_headers_by_ids(sheet_index, column_ids)
            df_name_with_selected_columns = f'{self.df_name}[{get_column_header_list_as_transpiled_code(column_headers)}]'
            df = self.df[column_headers]
        else:
            column_headers = df.columns.to_list()

        # The shorter code chunk is for dataframes that *don't* have any boolean columns
        # Boolean columns are a special case, because when we convert them to str then back
        # to bool, they all become True. So we have to convert them back to bool with a custom
        # function.
        code_chunk = [
            f'{df_name_with_selected_columns} = {df_name_with_selected_columns}.astype(str).replace("(?i){search_value}", "{replace_value}", regex=True).astype({df_name_with_selected_columns}.dtypes.to_dict())',
        ]

        if (any(df.dtypes == 'timedelta') and LooseVersion(pd.__version__) < LooseVersion("1.4")):
            raise MitoError(
                'version_error',
                'Pandas version error',
                'This version of pandas doesn\'t support replacing values in timedelta columns. Please upgrade to pandas 1.2 or later.',
            )
        if any(df.dtypes == bool):
            code_chunk = [
                f"non_bool_cols, bool_cols = {df_name_with_selected_columns}.select_dtypes(exclude='bool'), {df_name_with_selected_columns}.select_dtypes(include='bool')",
                f"{df_name}[non_bool_cols.columns] = non_bool_cols.astype(str).replace('(?i){search_value}', '{replace_value}', regex=True).astype(non_bool_cols.dtypes.to_dict())",
                f"{df_name}[bool_cols.columns] = bool_cols.astype(str).replace('(?i){search_value}', '{replace_value}', regex=True).applymap(cast_string_to_bool).astype(bool)",
            ]

        # Then, we always replace the search_value inside the column headers
        string_value_regex = re.compile(search_value, re.IGNORECASE)

        new_columns = [convert_to_original_type_or_str(re.sub(string_value_regex, replace_value, str(column)), type(column)) for column in column_headers]
        column_rename_map = dict(zip(column_headers, new_columns))
        column_rename_map = {k: v for k, v in column_rename_map.items() if k != v}
        if len(column_rename_map) > 0:
            code_chunk.append(f"{df_name}.rename(columns={get_column_header_map_as_code_string(column_rename_map)}, inplace=True)")
        return code_chunk, []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Exports the evaluate function, which takes a list of edit events
as well as the original dataframe, and returns the current state 
of the sheet as a dataframe
"""
import datetime
from distutils.version import LooseVersion
import re
import warnings
from typing import Any, List, Optional, Set, Tuple, Union

import pandas as pd

from mitosheet.column_headers import get_column_header_display
from mitosheet.errors import make_invalid_formula_error
from mitosheet.is_type_utils import (is_datetime_dtype,
                                                   is_number_dtype,
                                                   is_string_dtype)
from mitosheet.types import FORMULA_ENTIRE_COLUMN_TYPE, FrontendFormulaHeaderIndexReference, FrontendFormulaHeaderReference, IndexLabel
from mitosheet.transpiler.transpile_utils import (
    get_column_header_list_as_transpiled_code, get_column_header_as_transpiled_code)
from mitosheet.types import (ColumnHeader, FrontendFormula,
                             FormulaAppliedToType, RawParserMatch, ParserMatch,
                             ParserMatchSubstringRange, RowOffset)
from mitosheet.user.utils import get_pandas_version
from mitosheet.utils import is_prev_version

from mitosheet.array_utils import deduplicate_array

def is_quote(char: str) -> bool:
    """
    A helper to detect if a character is a quote
    """
    return char == '\'' or char == '"'

def get_string_matches(
        formula: str,
    ) -> List[ParserMatchSubstringRange]:
    """
    Returns a list of all the ranges that are strings within a formula,
    which is useful for other functions that do not want to change
    strings.

    Although column headers may contain quotes, when we check the overlap
    between the string matches and the column headers, those column headers
    that contain quotes will not be contained within the string, and therefor
    can still be detected as a valid column header.
    """
    if formula is None:
        return []

    # We find the ranges of the formula that are string constants;
    # we do not edit these! Taken from: https://stackoverflow.com/a/63707053/13113837
    string_iter_matches_double_quotes = re.finditer(r'"(?:(?:(?!(?<!\\)").)*)"', formula)
    string_iter_matches_single_quotes = re.finditer(r'\'(?:(?:(?!(?<!\\)\').)*)\'', formula)

    # Convert them to the match format
    string_matches_double_quotes = [(match.start(), match.end()) for match in string_iter_matches_double_quotes]
    string_matches_single_quotes = [(match.start(), match.end()) for match in string_iter_matches_single_quotes]

    return list(string_matches_double_quotes) + list(string_matches_single_quotes)

def remove_string_literals(formula: str) -> str:
    """
    Returns the formula with all of the string literals removed
    """
    string_matches = get_string_matches(formula)

    output_string = ""
    last_end = 0

    for match_start, match_end in string_matches:
        output_string += formula[last_end:match_start]
        last_end = match_end

    output_string += formula[last_end:]
    return output_string

def match_covered_by_matches( # type: ignore
        match_ranges: List[ParserMatchSubstringRange],
        match_range: ParserMatchSubstringRange
    ) -> bool:
    """
    Returns True iff a given match is contained by one
    of the matches in a string
    """
    start = match_range[0]
    end = match_range[1] # this is +1 after the last char of the string

    # We check if it's in any of the string ranges
    for other_match_range in match_ranges:
        string_start = other_match_range[0]
        string_end = other_match_range[1]

        if (string_start <= start and string_end >= end):
            return True

    return False

def safe_contains(
        formula: str, 
        substring: str,
        column_headers: List[ColumnHeader],
    ) -> bool:
    """
    Returns true if the formula contains substring. However, will not count
    the substring if it is inside of a string.

    If any of the column headers contain the substring, then we just skip
    this, as it's too hard of a check outwise.
    """
    for column_header in column_headers:
        if isinstance(column_header, str) and substring in column_header:
            return False

    # We do not search inside the string literals! 
    string_matches = get_string_matches(formula)

    for match in re.finditer(substring, formula):
        if not match_covered_by_matches(string_matches, (match.start(), match.end())):
            return True

    return False

def safe_contains_single_equals(formula: str, column_headers: List[ColumnHeader]) -> bool:
    """
        Check if the string contains a single equals sign that is not part of a substring:

        "A = B",                    # True
        "IF(A=B, 1, 0)",            # True
        'A==B',                     # False
        ' = ',                      # False
        "CONCAT(A, 'ABC=123')",     # False
        'A!=B',                     # False
        'A>=B',                     # False
        'A<=B',                     # False
    """

    # If the column headers contain an = then, we don't continue the check 
    # because it is too hard, and rare.
    for column_header in column_headers:
        if isinstance(column_header, str) and '=' in column_header:
            return False

    # We don't want to search inside of string literals, so we remove them
    formula = remove_string_literals(formula)

    strings_to_ignore = [
        '!=',
        '>=',
        '<=',
    ]

    # Remove all occurences of string_to_ignore from the formula. Do this outside of the regex
    # because its easy to do here, and the regex is already complicated enough
    for string_to_ignore in strings_to_ignore:
        formula = formula.replace(string_to_ignore, '')

    # This regex matches any single = that is not directly surrounded by another =
    pattern = r"(?<!\=)\=(?!\=)"
    return bool(re.search(pattern, formula))

def safe_contains_function(
        formula: str, 
        function: str,
        column_headers: List[ColumnHeader]
    ) -> bool:
    """
    Checks if a function is called in a formula. Returns False if the function
    is in a column header.
    """
    formula = formula.upper()
    function = function.upper()

    for column_header in column_headers:
        if isinstance(column_header, str) and function in column_header:
            return False

    # We do not search inside the string literals! 
    string_matches = get_string_matches(formula)

    for match in re.finditer(function, formula):
        if not match_covered_by_matches(string_matches, (match.start(), match.end())):
            # Check if this is a function
            end = match.end() # this is +1 after the last char of the string
            if end < len(formula) and formula[end] == '(':
                return True

    return False


def safe_count_function(formula: str, substring: str) -> int:
    """
    Counts the number of occurences of the substring in the formula
    not including the string ranges

    NOTE: is case insensitive!
    """
    formula = formula.upper()
    function = substring.upper()

    # We do not search inside the string literals! 
    string_matches = get_string_matches(formula)

    count = 0
    for match in re.finditer(function, formula):
        if not match_covered_by_matches(string_matches, (match.start(), match.end())):
            count += 1
    
    return count


def check_common_errors(
        formula: str,
        dfs: List[pd.DataFrame],
        df_names: List[str],
        sheet_index: int,
    ) -> None:
    """
    Helper function for checking a formula for common errors, for better
    communication with users. 

    If you want to throw a custom, specific error that is common for users 
    that can be detected when parsing a formula, add it here!
    """

    column_headers: List[ColumnHeader] = dfs[sheet_index].columns.to_list()

    if safe_contains(formula, "<>", column_headers):
        raise make_invalid_formula_error(
            formula,
            'Please use != instead of <> to check inequality.',
            error_modal=False
        )

    for sheet_name in df_names:
        if safe_contains(formula, f'{sheet_name}!', column_headers) and not safe_contains_function(formula, 'VLOOKUP', column_headers):
            raise make_invalid_formula_error(
                formula,
                f'Cross-sheet references are only allowed in calls to VLOOKUP',
                error_modal=False
            )

    # Remove leading white space from formula so we can easily remove
    # the leading = if it exists
    formula_clean = formula.lstrip()
    formula_without_leading_equals = formula_clean[1:] if formula_clean[0] == '=' else formula_clean
    if safe_contains_single_equals(formula_without_leading_equals, column_headers):
        raise make_invalid_formula_error(
            formula,
            'Use == instead of = to check equality.',
            error_modal=False
        )     

    # TODO: update this with vlookup formula in mind
    # If the user used a lookup formula, point them to merge instead!
    LOOKUP_FORMULAS = ['HLOOKUP', 'XLOOKUP']
    for lookup_formula in LOOKUP_FORMULAS:
        if safe_contains_function(formula.upper(), lookup_formula, column_headers):
            raise make_invalid_formula_error(
                formula,
                f'{lookup_formula} is not supported, but VLOOKUP is. See the documentation for more information.',
                error_modal=False
            )

    # If the parens are not matched, and no column header has a paren in it, then
    # we throw an error if we're sure that it's unmatched
    if safe_count_function(formula, r'\(') != safe_count_function(formula, r'\)'):
        for column_header in column_headers:
            if isinstance(column_headers, str):
                if '(' in column_header or ')' in column_header:
                    pass
        
        raise make_invalid_formula_error(
            formula,
            f'It looks like the parentheses are unmatched. Did you miss a )?',
            error_modal=False
        )

def is_no_index_after_column_header_match(formula: str, index: pd.Index, start: int, end: int) -> bool:
    # First, we check if it's an unqualified column header with no index. NOTE: this function
    # should only be called on a range where we already have a match on a column header,
    # and we're just checking that there's no index
    word_continues_before = (start - 1 >= 0 and formula[start - 1].isalnum())
    word_continues_after = end < len(formula) and formula[end].isalnum() 
    function_call = end < len(formula) and formula[end] == '('

    # We manually take care of instances where the index contains values that might not 
    # be alpha numeric, but are part of the index. So we allow . and - for decimals
    # and negative numbers, if it is not a range index that goes below 0
    if is_number_index(index):
        word_continues_after = word_continues_after or (end < len(formula) and (formula[end] == '.' or formula[end] == '-'))

    part_of_larger_item = word_continues_before or word_continues_after or function_call
    return not part_of_larger_item

def get_row_offset(index: pd.Index, formula_label: Union[str, bool, int, float, datetime.datetime], parsed_label: Any) -> Optional[RowOffset]:

    if formula_label == parsed_label:
        return 0

    parsed_label_index = index.get_indexer([parsed_label])[0]
    parsed_formula_label_index = index.get_indexer([formula_label])[0]

    # NOTE: edge case if -1 is a valid index, but not much to do
    if parsed_label_index != -1:
        row_offset: RowOffset = parsed_formula_label_index - parsed_label_index
        return row_offset

    return None

def is_number_index(index: pd.Index) -> bool:
    with warnings.catch_warnings():
        # Check pandas version is < 2.0
        if LooseVersion(pd.__version__) < LooseVersion('2.0.0'):
            warnings.simplefilter("ignore")
            if isinstance(index, pd.RangeIndex) or isinstance(index, pd.Int64Index) or isinstance(index, pd.UInt64Index) or isinstance(index, pd.Float64Index):
                return True
        else:
            index_dtype = str(index.dtype)
            return is_number_dtype(index_dtype)
    
    if is_number_dtype(str(index.dtype)):
        return True

    return False

def get_index_match_from_number_index(formula: str, formula_label: Union[str, bool, int, float], index: pd.Index, index_label_start: int) -> Optional[RawParserMatch]:
    if is_number_index(index):
        number_chars = ''
        for char in formula[index_label_start:]:
            # We allow it to start with a negative (must be first char)
            if len(number_chars) == 0 and char == '-':
                number_chars += char
            elif char.isnumeric() or char == '.':
                # OR, any character that is part of a number
                number_chars += char
            else:
                break
            
        if number_chars != '':
            if '.' in number_chars:
                parsed_label = float(number_chars)
            else:
                parsed_label = int(number_chars)

            row_offset = get_row_offset(index, formula_label, parsed_label)
            if row_offset is not None:
                return {
                    'type': '{INDEX}',
                    'substring_range': (index_label_start, index_label_start + len(number_chars)),
                    'unparsed': number_chars,
                    'parsed': parsed_label,
                    'row_offset': row_offset
                }
    return None

def is_datetime_index(index: pd.Index) -> bool:
    if isinstance(index, pd.DatetimeIndex):
        return True
    
    if is_datetime_dtype(str(index.dtype)):
        return True

    return False

def get_index_match_from_datetime_index(formula: str, formula_label: Union[str, bool, int, float], index: pd.Index, index_label_start: int) -> Optional[RawParserMatch]:
    if is_datetime_index(index):
        # The next characters should have the format '2007-01-22 00:00:00' (length 19 below)
        str_datetime = formula[index_label_start:index_label_start + 19]

        not_full_datetime = len(str_datetime) != 19 
        if not_full_datetime:
            return None

        try:
            parsed_datetime = datetime.datetime.strptime(str_datetime, '%Y-%m-%d %H:%M:%S')
            formula_label_parsed = formula_label if not isinstance(formula_label, str) else datetime.datetime.strptime(formula_label, '%Y-%m-%d %H:%M:%S')# We need to also parse this, as it's a string
            row_offset = get_row_offset(index, formula_label_parsed, parsed_datetime) #type: ignore
            if row_offset is not None:
                return {
                    'type': '{INDEX}',
                    'substring_range': (index_label_start, index_label_start + len(str_datetime)),
                    'unparsed': str_datetime,
                    'parsed': parsed_datetime,
                    'row_offset': row_offset
                }

        except ValueError:
            return None

    return None

def is_string_index(index: pd.Index) -> bool:
    return is_string_dtype(str(index.dtype))

def get_index_match_from_string_index(formula: str, formula_label: Union[str, bool, int, float], index: pd.Index, index_label_start: int) -> Optional[RawParserMatch]:
    if is_string_index(index):

        # Check for any string to the end of the string. We go backwards, to avoid issues
        # with index labels that are prefixes of each other
        for index_label_ends in reversed(range(index_label_start, len(formula) + 1)):
            potential_index_unparsed = formula[index_label_start:index_label_ends]
            row_offset = get_row_offset(index, formula_label, potential_index_unparsed)
            if row_offset is not None:
                return {
                    'type': '{INDEX}',
                    'substring_range': (index_label_start, index_label_start + len(potential_index_unparsed)),
                    'unparsed': potential_index_unparsed,
                    'parsed': potential_index_unparsed, # A parsed string is the same as an unparsed string
                    'row_offset': row_offset
                }

    return None


def get_raw_parser_matches(
        formula: str,
        formula_label: Union[str, bool, int, float], # Where the formula is written,
        string_matches: List[Any],
        dfs: List[pd.DataFrame],
        df_names: List[str],
        sheet_index: int
    ) -> List[RawParserMatch]:
    """
    Returns a list of RawParserMatch, which are either column header matches
    or matches on the index or a sheet. 

    The parsing strategy generally can be described as follows:
    1. Match columns headers that are not followed by indexes (for backwards compatibility)
    2. Match column headers that are followed by indexes
    3. Match sheet names that are followed by !

    (2) requires doing some reasoning about _what_ sort of index it is, 
    and therefore what sort of match is possible. Casting from the string
    to a parsed value is then done, and we can determine the row offset 
    of the match.

    formula = "=df_1!A0"
    formula_label = 1 (aka formula was written in B1)
    string_matches = []
    df = pd.DataFrame({'A': [1], 'B': [1]})

    We would get the matches: 
    {
        'type': '{SHEET}',
        'substring_range': (1, 4),
        'unparsed':'df_1',
        'parsed': 'df_1'
    }
    {
        'type': '{HEADER}',
        'substring_range': (1, 1),
        'unparsed':'A',
        'parsed': 'A',
        'row_offset': 0
    }
    {
        'type': '{INDEX}',
        'substring_range': (2, 2),
        'unparsed':'0',
        'parsed': 0,
        'row_offset': -1
    }
    """

    df = dfs[sheet_index]
    index = df.index
    column_headers: List[ColumnHeader] = df.columns.to_list()

    raw_parser_matches: List[RawParserMatch] = []

    for other_sheet_index, sheet_name in enumerate(df_names):
        if safe_contains(formula, f'{sheet_name}!', column_headers):
            for match_info in re.finditer(f'{sheet_name}!', formula):
                raw_parser_matches.append({
                    'type': '{SHEET}',
                    'substring_range': (match_info.start(), match_info.end()-1),
                    'unparsed': f'{sheet_name}!',
                    'parsed': sheet_name,
                    'row_offset': 0
                })

            # Update to look at the column headers in the other sheet
            column_headers = deduplicate_array(column_headers + dfs[other_sheet_index].columns.to_list())

    # We look for column headers from longest to shortest, to enable us
    # to issues if one column header is a substring of another
    # column header
    column_headers_sorted = sorted(column_headers, key=lambda ch: len(str(ch)), reverse=True)

    # First, we go through and replace all the column headers
    for column_header in column_headers_sorted:
        def find_column_headers(match: Any) -> Any:
            found_column_header: str = match.group() # type:ignore
            start = match.start()
            end = match.end()
            match_range = (start, end)

            # Do not replace the column header if it is in a string
            if match_covered_by_matches(string_matches, match_range):
                is_string = isinstance(column_header, str)
                starts_with_quote = is_quote(str(column_header)[0])
                ends_with_quote = is_quote(str(column_header)[-1])

                if is_string and not (starts_with_quote and ends_with_quote):
                    return found_column_header

            # If this column header was already covered by another column header
            # that has been found, then this column header is just a substring
            # of another column header, so we avoid matching it
            if match_covered_by_matches([match['substring_range'] for match in raw_parser_matches], match_range):
                return found_column_header

            # First, we check if it's an unqualified column header with no index
            if is_no_index_after_column_header_match(formula, index, start, end):
                raw_parser_matches.append({
                    'type': '{HEADER}',
                    'substring_range': match_range,
                    'parsed': column_header,
                    'unparsed': found_column_header,
                    'row_offset': 0
                })
                return

            # Second, check if column header is follwed by an index of any variety
            number_index_label_match = get_index_match_from_number_index(formula, formula_label, index, end)
            datetime_index_label_match = get_index_match_from_datetime_index(formula, formula_label, index, end)
            string_index_label_match = get_index_match_from_string_index(formula, formula_label, index, end)

            index_label_match = number_index_label_match or datetime_index_label_match or string_index_label_match or None
            if index_label_match is not None:
                raw_parser_matches.append({
                    'type': '{HEADER}',
                    'substring_range': match_range,
                    'parsed': column_header,
                    'unparsed': found_column_header,
                    'row_offset': index_label_match['row_offset']
                })
                raw_parser_matches.append(index_label_match)
                return 
            
            # NOTE: we add the column_header, not the found column header
            # as the found column header is a string, and the column_header 
            # may not be
            return found_column_header

        # First, just find all of the matches, without changing the string, as this shifts indexes
        # NOTE: for booleans, and for multi-index headers, we need to make the same transformation 
        # that we make on the frontend
        re.sub(re.escape(get_column_header_display(column_header)), find_column_headers, formula)

    # Sort the matches from start to end
    raw_parser_matches = sorted(raw_parser_matches, key=lambda x: x['substring_range'][0])

    return raw_parser_matches

def get_parser_matches(
        formula: str,
        formula_label: Union[str, bool, int, float], # Where the formula is written,
        string_matches: List[Any],
        dfs: List[pd.DataFrame],
        df_names: List[str],
        sheet_index: int
    ) -> List[ParserMatch]:
    """
    This function takes raw parser matches and turns them into ParserMatches, which are either:
    - {HEADER} which is a single header by itself (legacy)
    - {HEADER}{INDEX} which is how a user references a specific cell
    - {HEADER}:{HEADER} which is how a user references an entire column
    - {HEADER}{INDEX}:{HEADER}{INDEX} which is how a user references a range of specific cells
    - {SHEET}!{HEADER}:{HEADER} which is how a user references a range of columns in a separate sheet
    
    We take special care to match the longest options first, so that we don't match a {HEADER}{INDEX}
    when it is actual part of a {HEADER}{INDEX}:{HEADER}{INDEX}.

    This is also returned in reverse order, so that it is easy to work with.
    """
    

    def get_header_match(formula: str, raw_parser_matches: List[RawParserMatch], match_index: int) -> Optional[ParserMatch]:
        if match_index >= len(raw_parser_matches):
            return None

        curr_match = raw_parser_matches[match_index]
        if curr_match['type'] != '{HEADER}':
            return None

        is_final_solo_column_header = match_index + 1 >= len(raw_parser_matches) and curr_match['type'] == '{HEADER}'

        if is_final_solo_column_header:
            return {
                'type': '{HEADER}',
                'substring_range': curr_match['substring_range'],
                'unparsed': curr_match['unparsed'],
                'parsed': curr_match['parsed'],
                'row_offset': curr_match['row_offset']
            }
        
        next_match = raw_parser_matches[match_index + 1]
        not_followed_immediately_by_next_match = curr_match['substring_range'][1] != next_match['substring_range'][0]
        not_followed_by_colon_and_then_next_match = curr_match['substring_range'][1] != next_match['substring_range'][0] - 1 or formula[curr_match['substring_range'][1] + 1] != ':'

        if curr_match['type'] == '{HEADER}' and (not_followed_immediately_by_next_match or not_followed_by_colon_and_then_next_match):
            return {
                'type': '{HEADER}',
                'substring_range': curr_match['substring_range'],
                'unparsed': curr_match['unparsed'],
                'parsed': curr_match['parsed'],
                'row_offset': curr_match['row_offset']
            }
        
        return None
    
    def get_header_index_match(formula: str, raw_parser_matches: List[RawParserMatch], match_index: int) -> Optional[ParserMatch]:
        curr_match = raw_parser_matches[match_index]
        if match_index + 1 >= len(raw_parser_matches):
            return None

        next_match = raw_parser_matches[match_index + 1]
        followed_immediately_by_next_match = curr_match['substring_range'][1] == next_match['substring_range'][0]

        if curr_match['type'] == '{HEADER}' and followed_immediately_by_next_match and next_match['type'] == '{INDEX}':
            return {
                'type': '{HEADER}{INDEX}',
                'substring_range': (curr_match['substring_range'][0], next_match['substring_range'][1]),
                'unparsed': curr_match['unparsed'] + next_match['unparsed'],
                'parsed': (curr_match['parsed'], next_match['parsed']),
                'row_offset': curr_match['row_offset']
            }
        
        return None
    
    def get_header_header_match(formula: str, raw_parser_matches: List[RawParserMatch], match_index: int) -> Optional[ParserMatch]:
        curr_match = raw_parser_matches[match_index]
        if match_index + 1 >= len(raw_parser_matches):
            return None

        next_match = raw_parser_matches[match_index + 1]
        seperated_by_colon_from_next_match = curr_match['substring_range'][1] == next_match['substring_range'][0] - 1 and formula[curr_match['substring_range'][1]] == ':'

        if curr_match['type'] == '{HEADER}' and seperated_by_colon_from_next_match and next_match['type'] == '{HEADER}':
            return {
                'type': '{HEADER}:{HEADER}',
                'substring_range': (curr_match['substring_range'][0], next_match['substring_range'][1]),
                'unparsed': curr_match['unparsed'] + next_match['unparsed'],
                'parsed': (curr_match['parsed'], next_match['parsed']),
                'row_offset': curr_match['row_offset']
            }
        
        return None
    

    def get_sheet_header_header_match(formula: str, raw_parser_matches: List[RawParserMatch], match_index: int) -> Optional[ParserMatch]:
        curr_match = raw_parser_matches[match_index]
        if match_index + 2 >= len(raw_parser_matches):
            return None

        first_header_match = raw_parser_matches[match_index + 1]
        second_header_match = raw_parser_matches[match_index + 2]
        seperated_by_exclamation_from_next_match = curr_match['substring_range'][1] == first_header_match['substring_range'][0] - 1 and formula[curr_match['substring_range'][1]] == '!'
        seperated_by_colon_from_next_match = first_header_match['substring_range'][1] == second_header_match['substring_range'][0] - 1 and formula[first_header_match['substring_range'][1]] == ':'

        starts_with_sheet_reference = curr_match['type'] == '{SHEET}' and seperated_by_exclamation_from_next_match
        has_two_headers = first_header_match['type'] == '{HEADER}' and seperated_by_colon_from_next_match and second_header_match['type'] == '{HEADER}'

        if starts_with_sheet_reference and has_two_headers:
            return {
                'type': '{SHEET}!{HEADER}:{HEADER}',
                'substring_range': (curr_match['substring_range'][0], second_header_match['substring_range'][1]),
                'unparsed': curr_match['unparsed'] + first_header_match['unparsed'] + ':' + second_header_match['unparsed'],
                'parsed': (curr_match['parsed'], first_header_match['parsed'], second_header_match['parsed']),
                'row_offset': curr_match['row_offset']
            }
        
        return None

    def get_header_index_header_index_match(formula: str, raw_parser_matches: List[RawParserMatch], match_index: int) -> Optional[ParserMatch]:

        if match_index + 3 >= len(raw_parser_matches):
            return None

        first_specific_cell_match = get_header_index_match(formula, raw_parser_matches, match_index)
        second_specific_cell_match = get_header_index_match(formula, raw_parser_matches, match_index + 2)

        if first_specific_cell_match is None or second_specific_cell_match is None or isinstance(first_specific_cell_match['row_offset'], tuple) or isinstance(second_specific_cell_match['row_offset'], tuple):
            return None
        
        specific_cell_matches_seperated_by_colon = first_specific_cell_match['substring_range'][1] == second_specific_cell_match['substring_range'][0] - 1 and formula[first_specific_cell_match['substring_range'][1]] == ':'
        
        # Ensure the range is constructed with ascending index labels
        if first_specific_cell_match['row_offset'] >= second_specific_cell_match['row_offset']:
            start_of_range_specific_cell_match = first_specific_cell_match
            end_of_range_specific_cell_match = second_specific_cell_match  
        else:
            start_of_range_specific_cell_match = second_specific_cell_match
            end_of_range_specific_cell_match = first_specific_cell_match

        if specific_cell_matches_seperated_by_colon:
            return {
                'type': '{HEADER}{INDEX}:{HEADER}{INDEX}',
                'substring_range': (first_specific_cell_match['substring_range'][0], second_specific_cell_match['substring_range'][1]),
                'unparsed': first_specific_cell_match['unparsed'] + second_specific_cell_match['unparsed'],
                'parsed': (start_of_range_specific_cell_match['parsed'], end_of_range_specific_cell_match['parsed']),
                'row_offset': (start_of_range_specific_cell_match['row_offset'], end_of_range_specific_cell_match['row_offset']) # type: ignore
            }

        return None
     

    parser_matches: List[ParserMatch] = []

    raw_parser_matches = get_raw_parser_matches(
        formula,
        formula_label,
        string_matches,
        dfs,
        df_names,
        sheet_index
    )

    match_index = 0
    while match_index < len(raw_parser_matches):
        header_index_header_index_match = get_header_index_header_index_match(formula, raw_parser_matches, match_index)
        if header_index_header_index_match is not None:
            parser_matches.append(header_index_header_index_match)
            match_index += 4
            continue

        sheet_header_header_match = get_sheet_header_header_match(formula, raw_parser_matches, match_index)
        if sheet_header_header_match is not None:
            parser_matches.append(sheet_header_header_match)
            match_index += 3
            continue
        
        header_header_match = get_header_header_match(formula, raw_parser_matches, match_index)
        if header_header_match is not None:
            parser_matches.append(header_header_match)
            match_index += 2
            continue

        header_index_match = get_header_index_match(formula, raw_parser_matches, match_index)
        if header_index_match is not None:
            parser_matches.append(header_index_match)
            match_index += 2
            continue

        header_match = get_header_match(formula, raw_parser_matches, match_index)
        if header_match is not None:
            parser_matches.append(header_match)
            match_index += 1
            continue
        
        # If the user is trying to reference a sheet without it being {SHEET}{HEADER}{HEADER},
        # then we try to point them to using the correct syntax. 
        if raw_parser_matches[match_index]['type'] == '{SHEET}':
            raise make_invalid_formula_error(
                formula,
                f'Cross-sheet references are only allowed for ranges of columns.',
                error_modal=False
            )
            
        # If you've gotten here, then something has gone wrong
        raise make_invalid_formula_error(
            formula,
            'Something went wrong while parsing the formula. Please edit the formula and try again.',
            error_modal=False
        )

    # Put the parser matches in back-to-front order so we can work with them easily in consuming functions
    parser_matches.reverse()
    return parser_matches

                
def replace_sheet_names_and_column_headers_and_indexes(
        formula: str,
        formula_label: Union[str, bool, int, float],
        string_matches: List,
        dfs: List[pd.DataFrame],
        df_names: List[str],
        sheet_index: int
    ) -> Tuple[str, Set[ColumnHeader], Set[IndexLabel]]:
    """
    Returns a modified formula, where:
    1. column headers have been replaced with references to the dataframe
    2. index labels that follow those column headers are entirely removed

    If the index label referenced after a column header is not the the same as the
    formula_label, then we change the column header to have a .shift that shifts
    it the same relative amount. 
    """
    # We get all the matches first, from end to start order
    parser_matches = get_parser_matches(
        formula,
        formula_label,
        string_matches,
        dfs,
        df_names,
        sheet_index
    )

    df = dfs[sheet_index]
    df_name = df_names[sheet_index]

    column_headers = set()
    index_labels = set()

    def get_shift_string(column_dtype: str, row_offset: int) -> str:
        if row_offset == 0:
            return ''

        if is_number_dtype(column_dtype):
            return f'.shift({row_offset}, fill_value=0)'
        else:
            return f'.shift({row_offset})'
    
    def get_header_header_selection_code(column_header_one: ColumnHeader, column_header_two: ColumnHeader) -> str:
        # If the column headers are the same, we can generate much simpler code
        # TODO: we potentially can do better if the columnns are next to eachother as well
        if column_header_one == column_header_two:
            return f'[[{get_column_header_as_transpiled_code(column_header_two)}]]'
        else:
            return f'.loc[:, {get_column_header_as_transpiled_code(column_header_one)}:{get_column_header_as_transpiled_code(column_header_two)}]'
    
    # Then, go through from the end to the start, and actually replace all the column headers
    # and remove all of the index labels
    for match in parser_matches:
        match_type = match['type']
        start = match['substring_range'][0]
        end = match['substring_range'][1]

        replace_string = ''

        # We have to handle {HEADER}, {HEADER}{INDEX}, {HEADER}:{HEADER}, {HEADER}{INDEX}:{HEADER}{INDEX}, and {SHEET}!{HEADER}:{HEADER}
        if match_type == '{HEADER}':
            column_header = match['parsed']
            row_offset = match['row_offset']
            column_dtype = str(df[column_header].dtype)

            column_headers.add(column_header)

            transpiled_column_header = get_column_header_as_transpiled_code(column_header)
            replace_string = f'{df_name}[{transpiled_column_header}]{get_shift_string(column_dtype, row_offset)}' # type: ignore

        elif match_type == '{HEADER}{INDEX}':
            (column_header, index_label) = match['parsed']
            row_offset = match['row_offset']
            column_dtype = str(df[column_header].dtype)

            column_headers.add(column_header)
            index_labels.add(index_label)

            transpiled_column_header = get_column_header_as_transpiled_code(column_header)
            replace_string = f'{df_name}[{transpiled_column_header}]{get_shift_string(column_dtype, row_offset)}' # type: ignore
        
        elif match_type == '{SHEET}!{HEADER}:{HEADER}':
            (sheet_name, column_header_one, column_header_two) = match['parsed']
            cross_sheet_df = dfs[df_names.index(sheet_name)]
            # Ensure that the column headers are in the order they appear in the dataframe
            first_column_header, second_column_header = (column_header_one, column_header_two) if cross_sheet_df.columns.get_loc(column_header_one) <= cross_sheet_df.columns.get_loc(column_header_two) else (column_header_two, column_header_one) 

            # We add all of the column headers that are between these two headers
            cols = cross_sheet_df.loc[:, first_column_header:second_column_header].columns.tolist()
            column_headers.update(cols)
            replace_string = f'{sheet_name}{get_header_header_selection_code(first_column_header, second_column_header)}'

        elif match_type == '{HEADER}:{HEADER}':
            (column_header_one, column_header_two) = match['parsed']

            # Ensure that the column headers are in the order they appear in the dataframe
            first_column_header, second_column_header = (column_header_one, column_header_two) if df.columns.get_loc(column_header_one) <= df.columns.get_loc(column_header_two) else (column_header_two, column_header_one) 

            # We add all of the column headers that are between these two headers
            cols = df.loc[:, first_column_header:second_column_header].columns.tolist()
            column_headers.update(cols)
            
            replace_string = f'{df_name}{get_header_header_selection_code(first_column_header, second_column_header)}'

        elif match_type == '{HEADER}{INDEX}:{HEADER}{INDEX}':
            ((column_header_one, index_label_one), (column_header_two, index_label_two)) = match['parsed']
            (row_offset_one, row_offset_two) = match['row_offset'] # type: ignore

            # Ensure that the column headers are in the order they appear in the dataframe
            first_column_header, second_column_header = (column_header_one, column_header_two) if df.columns.get_loc(column_header_one) <= df.columns.get_loc(column_header_two) else (column_header_two, column_header_one)           
            
            # We add all of the column headers that are between these two headers
            cols = df.loc[:, first_column_header:second_column_header].columns.tolist()
            column_headers.update(cols)

            index_labels.add(index_label_one)
            index_labels.add(index_label_two)

            replace_string = f'RollingRange({df_name}{get_header_header_selection_code(first_column_header, second_column_header)}, {row_offset_one - row_offset_two + 1}, {-1 * row_offset_one})'
        
        formula = formula[:start] + replace_string + formula[end:]

    return formula, column_headers, index_labels


    
def replace_newlines_and_tabs(
        formula: str,
    ) -> str:
    """
    Removes all newlines and tabs that aren't in column headers
    """
    string_matches = get_string_matches(formula)

    def replace_newlines_and_tabs_internal(match):
        if not match_covered_by_matches(string_matches, (match.start(), match.end())):
            return ''
        else:
            return match.group()
    
    return re.sub(r'\n|\t', replace_newlines_and_tabs_internal, formula)
    

    
def replace_true_values(
        formula: str,
    ) -> str:
    """
    Replaces all values "TRUE" and "true" (that aren't in column headers) with True
    """
    string_matches = get_string_matches(formula)

    def replace_true_values_internal(match):
        if not match_covered_by_matches(string_matches, (match.start(), match.end())):
            return 'True'
        else:
            return match.group()
    return re.sub(r'TRUE|true', replace_true_values_internal, formula)
    

def replace_functions(
        formula: str,
    ) -> Tuple[str, Set[str]]:
    """
    Turns all the functions used in the formula upper case, and should
    be called after the column headers are replaced.
    """

    functions = set()
    # NOTE: as this function is called on the new formula with column headers replaced,
    # this detects string column headers, and in turn this means we don't replace inside
    # of column headers that could look like functions
    string_matches = get_string_matches(formula)

    def replace_functions_internal(match):

        word: str = match.group() # type:ignore
        end = match.end() # type:ignore # this is +1 after the last char of the string

        # If this is to_datetime or to_timedelta we skip this, as this is us casting to a datetime or 
        # timedetla for a column header. Similarly, we skip shift, as this is a function call b/c of a
        # row offset
        if word == 'to_datetime' or word == 'to_timedelta' or word == 'shift' or word == 'RollingRange':
            return word

        # Function
        if not match_covered_by_matches(string_matches, (match.start(), match.end())) and end < len(formula) and formula[end] == '(':
            # We turn all used functions into upper case in the translated Python
            # NOTE: this does not effect the original spreadsheet formula, which
            # may remain lower case. 
            function = word.upper()
            functions.add(function)

            return function
        else:
            return word
    
    # We match all words in formula, and send them through the replace function.
    # See documentation here: https://docs.python.org/3/library/re.html#re.sub
    formula_with_functions = re.sub(r'\w+', replace_functions_internal, formula)

    return formula_with_functions, functions


def parse_formula(
        formula: Optional[str], 
        column_header: ColumnHeader, 
        formula_label: Union[str, bool, int, float],
        index_labels_formula_is_applied_to: FormulaAppliedToType,
        dfs: List[pd.DataFrame],
        df_names: List[str],
        sheet_index: int,
        include_df_set: bool=True,
    ) -> Tuple[str, Set[str], Set[ColumnHeader], Set[IndexLabel]]:
    """
    Returns a representation of the formula that is easy to handle, specifically
    by returning (python_code, functions, column_header_dependencies), where column_headers
    is a list of dependencies that the formula references.

    If include_df_set, then will return {df_name}[{column_header}] = {parsed formula}, and if
    not then will just return {parsed formula}
    """
    df = dfs[sheet_index]
    df_name = df_names[sheet_index]

    # If the column doesn't have a formula, then there are no dependencies, duh!
    if formula is None or formula == '':
        return '', set(), set(), set()

    check_common_errors(formula, dfs, df_names, sheet_index)

    # Chop off any whitespace at the start
    formula = formula.lstrip()

    # Chop off the =, if it exists. We also accept formulas
    # that don't have an equals
    if formula.startswith('='):
        formula = formula[1:]

    # First, we get the string matches
    string_matches = get_string_matches(formula)

    # Then, we get the column header matches, as well as replace them with valid python
    code_with_column_headers, column_header_dependencies, index_label_dependencies = replace_sheet_names_and_column_headers_and_indexes(
        formula, 
        formula_label,
        string_matches,
        dfs,
        df_names,
        sheet_index
    )

    code_without_newlines_or_tabs = replace_newlines_and_tabs(code_with_column_headers)
    code_with_values_replaced = replace_true_values(code_without_newlines_or_tabs)
    code_with_functions, functions = replace_functions(code_with_values_replaced)

    transpiled_column_header = get_column_header_as_transpiled_code(column_header)

    if include_df_set:
        if index_labels_formula_is_applied_to['type'] == FORMULA_ENTIRE_COLUMN_TYPE:
            final_code = f'{df_name}[{transpiled_column_header}] = {code_with_functions}'
        else:
            index_labels = index_labels_formula_is_applied_to["index_labels"] # type: ignore
            # If you're writing to an index that is a datetime, we make sure that these objects are datetimes (this is only necessary)
            # on versions of pandas earlier than 1.0
            if is_datetime_index(df.index) and is_prev_version(get_pandas_version(), '1.0.0'):
                index_labels = pd.to_datetime(index_labels)

            final_code = f'{df_name}.loc[{get_column_header_list_as_transpiled_code(index_labels)}, [{transpiled_column_header}]] = {code_with_functions}' # type: ignore

    else:
        final_code = f'{code_with_functions}'

    return final_code, functions, column_header_dependencies, index_label_dependencies


def get_frontend_formula_header_index_reference(
        column_header: ColumnHeader,
        row_offset: int,
    ) -> FrontendFormulaHeaderIndexReference:
    return {
        'type': '{HEADER}{INDEX}',
        'display_column_header': get_column_header_display(column_header),
        'row_offset': row_offset,
    }

def get_frontend_formula_header_reference(
        column_header: ColumnHeader,
    ) -> FrontendFormulaHeaderReference:
    return {
        'type': '{HEADER}',
        'display_column_header': get_column_header_display(column_header),
    }


def get_frontend_formula(
    formula: Optional[str], 
    formula_label: Union[str, bool, int, float],
    dfs: List[pd.DataFrame],
    df_names: List[str],
    sheet_index: int
) -> FrontendFormula:
    # Returns a formula string in a representation that the frontend can correctly update
    # when the user views it in a different cell

    # If the column doesn't have a formula, then there are no dependencies, duh!
    if formula is None or formula == '':
        return []

    parser_matches = get_parser_matches(
        formula,
        formula_label,
        get_string_matches(formula),
        dfs,
        df_names,
        sheet_index
    )
        
    frontend_formula: FrontendFormula = []
    start = 0
    for match in reversed(parser_matches):
        match_type = match['type']
        parser_match_start = match['substring_range'][0]
        parser_match_end = match['substring_range'][1]

        frontend_string_part = formula[start: parser_match_start]

        if len(frontend_string_part) > 0:
            frontend_formula.append({'type': 'string part', 'string': frontend_string_part})

        if match_type == '{HEADER}':
            column_header = match['parsed']
            row_offset: int = match['row_offset'] # type: ignore
            frontend_formula.append(get_frontend_formula_header_index_reference(column_header, row_offset))

        elif match_type == '{HEADER}{INDEX}':
            column_header, _ = match['parsed']
            row_offset: int = match['row_offset'] # type: ignore
            frontend_formula.append(get_frontend_formula_header_index_reference(column_header, row_offset))

        elif match_type == '{HEADER}:{HEADER}':
            (column_header_one, column_header_two) = match['parsed']
            frontend_formula.append(get_frontend_formula_header_reference(column_header_one))
            frontend_formula.append({'type': 'string part', 'string': ':'})
            frontend_formula.append(get_frontend_formula_header_reference(column_header_two))

        elif match_type == '{HEADER}{INDEX}:{HEADER}{INDEX}':
            ((column_header_one, _), (column_header_two, _)) = match['parsed']
            (row_offset_one, row_offset_two) = match['row_offset'] # type: ignore
            frontend_formula.append(get_frontend_formula_header_index_reference(column_header_one, row_offset_one))
            frontend_formula.append({'type': 'string part', 'string': ':'})
            frontend_formula.append(get_frontend_formula_header_index_reference(column_header_two, row_offset_two))

        elif match_type == '{SHEET}!{HEADER}:{HEADER}':
            (sheet_name, column_header_one, column_header_two) = match['parsed']
            frontend_formula.append({ 'type': '{SHEET}', 'display_sheet_name': sheet_name})
            frontend_formula.append({'type': 'string part', 'string': '!'})
            frontend_formula.append(get_frontend_formula_header_reference(column_header_one))
            frontend_formula.append({'type': 'string part', 'string': ':'})
            frontend_formula.append(get_frontend_formula_header_reference(column_header_two))


        start = parser_match_end 

    # Make sure we get the rest of the formula
    frontend_string_part = formula[start: len(formula)]
    if len(frontend_string_part) > 0:
        frontend_formula.append({
            'type': 'string part',
            'string': frontend_string_part
        })

    return frontend_formula


def get_backend_formula_from_frontend_formula(
    frontend_formula: FrontendFormula,
    formula_label: Union[str, bool, int, float],
    df: pd.DataFrame,
) -> str:
    # This function is useful on the backend as a testing utility, so that we can make sure
    # the FrontendFormula can reconstruct the backend formula properly
    formula = ''
    for formula_part in frontend_formula:
        if formula_part['type'] == 'string part':
            formula += formula_part['string'] # type: ignore
        elif formula_part['type'] == '{HEADER}':
            formula += formula_part['display_column_header']
        elif formula_part['type'] == '{SHEET}':
            formula += formula_part['display_sheet_name'] # type: ignore
        else:
            formula += formula_part['display_column_header'] # type: ignore
            formula += get_column_header_display(df.index[df.index.get_indexer([formula_label])[0] - formula_part['row_offset']]) # type: ignore
    return formula



from datetime import datetime, timedelta
from typing import Optional, Union

import pandas as pd
from mitosheet.is_type_utils import is_string_dtype

from mitosheet.public.v1.sheet_functions.types.utils import get_to_datetime_params


def cast_string_to_datetime(
        s: str,
    ) -> Optional[Union[pd.Timestamp, datetime]]:

    to_datetime_params = get_to_datetime_params(pd.Series(s))

    result = pd.to_datetime(
        s,
        errors='coerce',
        **to_datetime_params
    ) 
    
    # Make sure we don't return a NaT
    if pd.isnull(result):
        return None
    else:
        return result # type: ignore

def cast_to_datetime(unknown: Union[str, int, float, bool, datetime, timedelta]) -> Optional[Union[datetime, pd.Timestamp]]:
    if isinstance(unknown, str):
        return cast_string_to_datetime(unknown)
    elif isinstance(unknown, int):
        return datetime.fromtimestamp(unknown)
    elif isinstance(unknown, float):
        return datetime.fromtimestamp(unknown)
    elif isinstance(unknown, datetime) or isinstance(unknown, pd.Timestamp):
        return unknown

    return None


def cast_series_to_datetime(series: pd.Series) -> pd.Series:
    """
    String casting can be very slow if you do it element-wise. As such, we provide a 
    function that casts the the entire series at once. 
    """

    dtype = str(series.dtype)
    if is_string_dtype(dtype):
        return pd.to_datetime(
            series,
            errors='coerce',
            **get_to_datetime_params(series)
        ) 
    
    # Otherwise, we just cast it element-wise
    return series.apply(cast_to_datetime)




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from copy import copy
from typing import List, Optional, TYPE_CHECKING, Any, Dict, Tuple
from mitosheet.state import State

from mitosheet.types import ColumnID, ColumnHeader
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.transpiler.transpile_utils import \
    get_column_header_list_as_transpiled_code
from mitosheet.code_chunks.step_performers.column_steps.reorder_column_code_chunk import ReorderColumnCodeChunk
from mitosheet.code_chunks.no_op_code_chunk import NoOpCodeChunk

if TYPE_CHECKING:
    from mitosheet.code_chunks.step_performers.column_steps.add_column_code_chunk import AddColumnCodeChunk
    from mitosheet.code_chunks.step_performers.column_steps.rename_columns_code_chunk import RenameColumnsCodeChunk
else:
    AddColumnCodeChunk = Any


class DeleteColumnsCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, sheet_index: int, column_ids: List[ColumnID]):
        super().__init__(prev_state)
        self.sheet_index = sheet_index
        self.column_ids = column_ids

        self.df_name = self.prev_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'Deleted columns'
    
    def get_description_comment(self) -> str:
        column_headers = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index, self.column_ids)
        return f'Deleted columns {", ".join([str(ch) for ch in column_headers])}'

    def get_code(self) -> Tuple[List[str], List[str]]:

        column_headers_list_string = get_column_header_list_as_transpiled_code(
            [self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, column_id) for column_id in self.column_ids]
        )

        return [f'{self.df_name}.drop({column_headers_list_string}, axis=1, inplace=True)'], []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]

    def _combine_right_with_delete_columns_code_chunk(self, other_code_chunk: "DeleteColumnsCodeChunk") -> Optional["DeleteColumnsCodeChunk"]:
        if not self.params_match(other_code_chunk, ['sheet_index']):
            return None

        first_column_ids = self.column_ids
        second_column_ids = other_code_chunk.column_ids
        all_column_ids = first_column_ids + second_column_ids

        # Use a loop rather than a set so we preserve the order of the columns being deleted
        new_column_ids = []
        for column_id in all_column_ids:
            if column_id not in new_column_ids:
                new_column_ids.append(column_id)

        return DeleteColumnsCodeChunk(
            self.prev_state,
            self.sheet_index,
            new_column_ids
        )

    def combine_right(self, other_code_chunk: CodeChunk) -> Optional[CodeChunk]:
        if isinstance(other_code_chunk, DeleteColumnsCodeChunk):
            return self._combine_right_with_delete_columns_code_chunk(other_code_chunk)
            
        return None

    def _combine_left_reorder_column_code_chunk(self, other_code_chunk: ReorderColumnCodeChunk) -> Optional["CodeChunk"]:
        if not self.params_match(other_code_chunk, ['sheet_index']):
            return None

        reordered_column_id = other_code_chunk.column_id
        column_ids = self.column_ids

        if reordered_column_id in column_ids:
            return DeleteColumnsCodeChunk(
                other_code_chunk.prev_state,
                self.sheet_index,
                self.column_ids
            )

        return None

    def _combine_left_add_column_code_chunk(self, add_column_code_chunk: "AddColumnCodeChunk") -> Optional["CodeChunk"]:
        if not self.params_match(add_column_code_chunk, ['sheet_index']):
            return None

        added_column_id = add_column_code_chunk.new_column_id

        if added_column_id in self.column_ids:

            # Remove the added column, as we can just skip and no longer need to delete
            new_column_ids = copy(self.column_ids)
            new_column_ids.remove(added_column_id)

            # If there's nothing new, then we return a noop
            if len(new_column_ids) == 0:
                return NoOpCodeChunk(add_column_code_chunk.prev_state)
            else:
                # Otherwise, just delete what else is deleted in this step
                return DeleteColumnsCodeChunk(
                    add_column_code_chunk.prev_state,
                    self.sheet_index,
                    new_column_ids
                )

        return None

    def _combine_left_rename_columns_code_chunk(self, other_code_chunk: "RenameColumnsCodeChunk") -> Optional["CodeChunk"]:
        if not self.params_match(other_code_chunk, ['sheet_index']):
            return None
        
        column_ids_to_new_column_headers = other_code_chunk.column_ids_to_new_column_headers
        deleted_column_ids: List[ColumnID] = self.column_ids

        to_remove_from_rename = []
        for column_id in deleted_column_ids:
            if column_id in column_ids_to_new_column_headers:
                to_remove_from_rename.append(column_id)

        # If there is not any overlap between renamed and deleted, we change nothing
        if len(to_remove_from_rename) == 0:
            return None

        # If we are removing all of the renames, then we can just skip the renames, and 
        # just do the deletes
        if len(to_remove_from_rename) == len(column_ids_to_new_column_headers):    
            return DeleteColumnsCodeChunk(
                other_code_chunk.prev_state,
                self.sheet_index,
                self.column_ids
            )
        
        # Otherwise, we don't optimize
        return None
        
    def combine_left(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        if isinstance(other_code_chunk, ReorderColumnCodeChunk):
            return self._combine_left_reorder_column_code_chunk(other_code_chunk)
        from mitosheet.code_chunks.step_performers.column_steps.add_column_code_chunk import AddColumnCodeChunk
        if isinstance(other_code_chunk, AddColumnCodeChunk):
            return self._combine_left_add_column_code_chunk(other_code_chunk)
        from mitosheet.code_chunks.step_performers.column_steps.rename_columns_code_chunk import RenameColumnsCodeChunk
        if isinstance(other_code_chunk, RenameColumnsCodeChunk):
            return self._combine_left_rename_columns_code_chunk(other_code_chunk)

        return None



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import traceback
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd

from mitosheet.ai.recon import update_state_by_reconing_dataframes
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.user_defined_edit_code_chunk import \
    UserDefinedEditCodeChunk
from mitosheet.errors import MitoError
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.user_defined_function_utils import \
    get_user_defined_function_param_type_and_execute_value_and_transpile_value
from mitosheet.step_performers.utils.utils import get_param


class UserDefinedEditStepPerformer(StepPerformer):
    """
    Allows the user to make an edit to a dataframe with with a user 
    defined function.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'user_defined_edit'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:

        edit_name: str = get_param(params, 'edit_name')
        edit_params: Any = get_param(params, 'edit_params')
        
        # We make a new state to modify it
        post_state = prev_state.copy() # TODO: update the deep copies

        pandas_start_time = perf_counter()

        try:
            edit_function = next(f for f in post_state.user_defined_editors if f.__name__ == edit_name)
        except:
            raise MitoError(
                'user_defined_importer_not_found',
                f"Importer {edit_name} not found.",
                f"User defined importer {edit_name} not found. Please check that it is defined in the `editors` list passed to mitosheet.sheet.",
                error_modal=True
            )
        
        user_defined_function_params = get_user_defined_function_param_type_and_execute_value_and_transpile_value(post_state, edit_function, edit_params)
    
        try:
            result = edit_function(**{param_name: execute_param for param_name, (_, execute_param, _) in user_defined_function_params.items()})
        except:
            traceback_final_line = traceback.format_exc().splitlines()[-1]

            raise MitoError(
                'user_defined_editor_error',
                f"Editor {edit_name} raised an error.",
                f"User defined editor {edit_name} raised an error: {traceback_final_line}",
                error_modal=False
            )

        if not isinstance(result, pd.DataFrame):
            raise MitoError(
                'user_defined_editor_error',
                f"Editor {edit_name} raised an error.",
                f"User defined editor {edit_name} must return a single pandas dataframe. Instead it returned a result of type {type(result)}",
                error_modal=False
            )
        
        # Find the single dataframe param - so that we can get the sheet index for transpilation and updating the state
        df_names = {transpiled_param for _, (param_type, _, transpiled_param) in user_defined_function_params.items() if param_type == 'DataFrame'}
        if len(df_names) != 1:
            raise ValueError(
                f'Please ensure the editor {edit_name} takes a single dataframe as input. Otherwise, Mito cannot detect which dataframe you are editing.'
            )
        
        sheet_index = post_state.df_names.index(df_names.pop())
        post_state, _ = update_state_by_reconing_dataframes(post_state, sheet_index, post_state.dfs[sheet_index], result)

        pandas_processing_time = perf_counter() - pandas_start_time

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'sheet_index': sheet_index,
            'user_defined_function_params': user_defined_function_params,
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            UserDefinedEditCodeChunk(
                prev_state, 
                get_param(params, 'edit_name'),
                get_param(execution_data if execution_data is not None else dict(), 'sheet_index'),
                get_param(execution_data if execution_data is not None else dict(), 'user_defined_function_params')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        # TODO: we could improve this ideally, but for now we return everything
        return set()
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List

from mitosheet.step_performers.graph_steps.plotly_express_graphs import DO_NOT_CHANGE_PAPER_BGCOLOR_DEFAULT, DO_NOT_CHANGE_PLOT_BGCOLOR_DEFAULT, DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT

def upgrade_graph_1_to_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Adds the default styling params 

    OLD: 
        'step_version': 1,
        'step_type': 'graph',
        'params: {
            graph_id: GraphID
            graph_preprocessing: {
                safety_filter_turned_on_by_user: boolean
            },
            graph_creation: {
                graph_type: GraphType,
                sheet_index: int
                x_axis_column_ids: ColumnID[],
                y_axis_column_ids: ColumnID[],
                color: columnID: columnID
            },
            graph_styling: {},
            graph_rendering: {
                height: int representing the div width
                width: int representing the div width
            }
        }
    }

    NEW: 
        'step_version': 2,
        'step_type': 'graph',
        'params: {
            graph_id: GraphID
            graph_preprocessing: {
                safety_filter_turned_on_by_user: boolean
            },
            graph_creation: {
                graph_type: GraphType,
                sheet_index: int
                x_axis_column_ids: ColumnID[],
                y_axis_column_ids: ColumnID[],
                color: columnID: columnID
            },
            graph_styling: {
                title: {
                    title: string | None
                    visible: boolean
                },
                xaxis: {
                    title: string | None,
                    visible: boolean,
                    rangeslider: {
                        visible: boolean
                    }
                },
                yaxis: {
                    title: string | None,
                    visible: boolean
                },
                showlegend: boolean
            },
            graph_rendering: {
                height: int representing the div width
                width: int representing the div width
            }
        }
    }
    """
    params = step['params']
    params['graph_styling'] = {
        'title': {
            # Note: we don't add the title because its set to None and therefore doesn't appear as a param in the step
            'visible': True
        },
        'xaxis': {
            # Note: we don't add the title because its set to None and therefore doesn't appear as a param in the step
            'visible': True,
            'rangeslider': {
                'visible': True
            }
        },
        'yaxis': {
            # Note: we don't add the title because its set to None and therefore doesn't appear as a param in the step
            'visible': True
        },
        'showlegend': True
    }

    return [{
        "step_version": 2, 
        "step_type": "graph", 
        "params": params
    }] + later_steps


def upgrade_graph_2_to_3(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Adds paper_bgcolor and plot_bgcolor to the graph styling params, as well
    as title_font_color for all the axes

    OLD: 
        'step_version': 2,
        'step_type': 'graph',
        'params: {
            graph_id: GraphID
            graph_preprocessing: {
                safety_filter_turned_on_by_user: boolean
            },
            graph_creation: {
                graph_type: GraphType,
                sheet_index: int
                x_axis_column_ids: ColumnID[],
                y_axis_column_ids: ColumnID[],
                color: columnID: columnID
            },
            graph_styling: {
                title: {
                    title: string | None
                    visible: boolean
                },
                xaxis: {
                    title: string | None,
                    visible: boolean,
                    rangeslider: {
                        visible: boolean
                    }
                },
                yaxis: {
                    title: string | None,
                    visible: boolean
                },
                showlegend: boolean
            },
            graph_rendering: {
                height: int representing the div width
                width: int representing the div width
            }
        }
    }

    NEW: 
        'step_version': 3,
        'step_type': 'graph',
        'params: {
            graph_id: GraphID
            graph_preprocessing: {
                safety_filter_turned_on_by_user: boolean
            },
            graph_creation: {
                graph_type: GraphType,
                sheet_index: int
                x_axis_column_ids: ColumnID[],
                y_axis_column_ids: ColumnID[],
                color: columnID: columnID
            },
            graph_styling: {
                title: {
                    title: string | None
                    visible: boolean,
                    title_font_color: string
                },
                xaxis: {
                    title: string | None,
                    visible: boolean,
                    title_font_color: string
                    rangeslider: {
                        visible: boolean
                    }
                },
                yaxis: {
                    title: string | None,
                    visible: boolean,
                    title_font_color: string
                },
                showlegend: boolean,
                paper_bgcolor: string,
                plot_bgcolor: string
            },
            graph_rendering: {
                height: int representing the div width
                width: int representing the div width
            }
        }
    }
    """
    params = step['params']
    
    params['graph_styling']['paper_bgcolor'] = DO_NOT_CHANGE_PAPER_BGCOLOR_DEFAULT
    params['graph_styling']['plot_bgcolor'] = DO_NOT_CHANGE_PLOT_BGCOLOR_DEFAULT

    params['graph_styling']['title']['title_font_color'] = DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT
    params['graph_styling']['xaxis']['title_font_color'] = DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT
    params['graph_styling']['yaxis']['title_font_color'] = DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT

    return [{
        "step_version": 3, 
        "step_type": "graph", 
        "params": params
    }] + later_steps



def upgrade_graph_3_to_4(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Adds required graph styling params:
    - xaxis: showgrid
    - yaxis: showgrid
    - legend: {
            orientation: 'v' | 'h'
        }

    OLD: 
        'step_version': 3,
        'step_type': 'graph',
        'params: {
            graph_id: GraphID
            graph_preprocessing: {
                safety_filter_turned_on_by_user: boolean
            },
            graph_creation: {
                graph_type: GraphType,
                sheet_index: int
                x_axis_column_ids: ColumnID[],
                y_axis_column_ids: ColumnID[],
                color: columnID: columnID
            },
            graph_styling: {
                title: {
                    title: string | None
                    visible: boolean,
                    title_font_color: string
                },
                xaxis: {
                    title: string | None,
                    visible: boolean,
                    title_font_color: string
                    rangeslider: {
                        visible: boolean
                    }
                },
                yaxis: {
                    title: string | None,
                    visible: boolean,
                    title_font_color: string
                },
                showlegend: boolean,
                paper_bgcolor: string,
                plot_bgcolor: string
            },
            graph_rendering: {
                height: int representing the div width
                width: int representing the div width
            }
        }
    }

    NEW: 
        'step_version': 4,
        'step_type': 'graph',
        'params: {
            graph_id: GraphID
            graph_preprocessing: {
                safety_filter_turned_on_by_user: boolean
            },
            graph_creation: {
                graph_type: GraphType,
                sheet_index: int
                x_axis_column_ids: ColumnID[],
                y_axis_column_ids: ColumnID[],
                color: columnID: columnID
            },
            graph_styling: {
                title: {
                    title: string | None
                    visible: boolean,
                    title_font_color: string
                },
                xaxis: {
                    title: string | None,
                    visible: boolean,
                    title_font_color: string
                    showgrid: boolean
                    rangeslider: {
                        visible: boolean
                    }
                },
                yaxis: {
                    title: string | None,
                    visible: boolean,
                    title_font_color: string
                    showgrid: boolean
                },
                showlegend: boolean,
                legend: {
                    orientation: 'v' | 'h'
                }
                paper_bgcolor: string,
                plot_bgcolor: string
            },
            graph_rendering: {
                height: int representing the div width
                width: int representing the div width
            }
        }
    }
    """
    params = step['params']
    params['graph_styling']['xaxis']['showgrid'] = True
    params['graph_styling']['yaxis']['showgrid'] = True
    params['graph_styling']['legend'] = dict()
    params['graph_styling']['legend']['orientation'] = 'v'

    return [{
        "step_version": 4, 
        "step_type": "graph", 
        "params": params
    }] + later_steps


import json
import os
import sys
from mitosheet.telemetry.telemetry_utils import log
import analytics

MITO_STARTER_NOTEBOOK_PATH = './mito-starter-notebook.ipynb'
MITO_STARTER_NOTEBOOK_CONTENTS = {
 "cells": [
  {
   "cell_type": "code",
   "execution_count": None,
   "id": "c0957234-ba57-4199-adfa-9d17ede0b8f9",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Run this cell to render a mitosheet\n",
    "# See mito's documentation here: https://docs.trymito.io/how-to/creating-a-mitosheet\n",
    "# Join our Discord for support here: https://discord.com/invite/XdJSZyejJU\n"
    "\n",
    "import mitosheet\n",
    "mitosheet.sheet()"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.5"
  },
  "widgets": {
   "application/vnd.jupyter.widget-state+json": {
    "state": {},
    "version_major": 2,
    "version_minor": 0
   }
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}

def try_create_starter_notebook() -> bool:
    """
    Creates MITO_STARTER_NOTEBOOK_CONTENTS at MITO_STARTER_NOTEBOOK_PATH,
    which helps the user get started.

    If the file already exists, will not replace it, as to not overwrite
    users work in this area.
    """
    try:
        if not os.path.exists(MITO_STARTER_NOTEBOOK_PATH):
            with open(MITO_STARTER_NOTEBOOK_PATH, 'w+') as f:
                f.write(json.dumps(MITO_STARTER_NOTEBOOK_CONTENTS))
        log('hello_command_create_starter_notebook')
        return True
    except Exception as e:
        log(
            'hello_command_create_starter_notebook_error', 
            failed=True,
            error=e
        )
        return False



def run_hello() -> None:
    log('hello_command')

    print()
    print()
    print("\t 👋 Welcome to Mito!")
    print()
    print("\tThis command launches Mito in JupyterLab, but Mito is also available in Jupyter Notebook Classic, Streamlit, and Dash.")
    print()
    print("\tSee our docs here: https://docs.trymito.io")
    print()
    print()

    try_create_starter_notebook()

    # Flush any logs
    analytics.flush()

    # Launch JupyterLab
    try:
        os.execl(sys.executable, 'python', '-m', 'jupyter', 'lab', MITO_STARTER_NOTEBOOK_PATH)
    except Exception as e:
        log('hello_command_start_jupyterlab_error', error=e, failed=True)

    # Flush any logs, again
    analytics.flush()

    

    


import json
import time
from queue import Queue
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pandas as pd

from mitosheet.mito_backend import MitoBackend
from mitosheet.selection_utils import get_selected_element
from mitosheet.streamlit.v1 import RunnableAnalysis
from mitosheet.types import (CodeOptions, MitoFrontendIndexAndSelections,
                             MitoTheme, ParamMetadata)
from mitosheet.utils import get_new_id


class SpreadsheetResult():

    def __init__(
        self, 
        dfs: List[pd.DataFrame],
        code: List[str],
        fully_parameterized_function: str,
        param_metadata: List[ParamMetadata],
        code_options: CodeOptions,
        index_and_selections: Optional[MitoFrontendIndexAndSelections]=None
    ):
        self.__dfs = dfs
        self.__code = code
        self.__index_and_selections = index_and_selections
        self.__fully_parameterized_function = fully_parameterized_function
        self.__param_metadata = param_metadata
        self.__code_options = code_options

    def dfs(self) -> List[pd.DataFrame]:
        return self.__dfs
    
    def code(self) -> str:
        return "\n".join(self.__code)
    
    def selection(self) -> Optional[Union[pd.DataFrame, pd.Series]]:
        return get_selected_element(self.__dfs, self.__index_and_selections)
    
    def analysis(self) -> RunnableAnalysis:
        return RunnableAnalysis(self.code(), self.__code_options, self.__fully_parameterized_function, self.__param_metadata)
    
WRONG_CALLBACK_ERROR_MESSAGE = """Error: Registering a callback with an Input or State referencing a Mito Spreadsheet requires using the @mito_callback decorator, rather than the @callback decorator.

To get the proper value from the {prop_name}, change your callback to use the @mito_callback decorator instead of the @callback decorator.

See more: https://docs.trymito.io/mito-for-dash/api-reference#callback-props-and-types

{num_messages}
{prop_name}
{id}
{session_key}"""

ID_TYPE = 'spreadsheet'
         
try:
    from dash.development.base_component import Component

    from dash import Input, Output, State, callback


    class Spreadsheet(Component):
        
        # See documentation in the get_instance method
        instances: Dict[str, Tuple[Any, Dict[str, Any]]] = dict()

        _children_props: List[str] = []
        _base_nodes = ['children']
        _namespace = 'dash_spreadsheet_v1'
        _type = 'MitoDashWrapper'
        _prop_names = ['id', 'all_json', 'data', 'import_folder', 'spreadsheet_result', 'spreadsheet_selection', 'mito_id', 'session_key']
        _valid_wildcard_attributes: List[str] = []
        available_properties = ['id', 'all_json', 'data', 'import_folder']
        available_wildcard_properties: List[str] = []

        def __init__(
            self, 
            *args: Union[pd.DataFrame, str, None],
            id: Dict[str, str],
            import_folder: Optional[str]=None,
            code_options: Optional[CodeOptions]=None,
            df_names: Optional[List[str]]=None,
            sheet_functions: Optional[List[Callable]]=None, 
            importers: Optional[List[Callable]]=None,
            editors: Optional[List[Callable]]=None,
            theme: Optional[MitoTheme]=None,
            track_selection: bool=False,
            mito_frontend_key: Optional[str]=None
        ):    
            
            # First, we check that the user has passed an ID in the appropriate format
            # as this is required for the component to work. It must be a dict with the 
            # keys type and id, where type == ID_TYPE and id is a unique string
            if not isinstance(id, dict):
                raise Exception("You must pass an ID to the Spreadsheet component as a dict with the keys type and id")
            if 'type' not in id:
                raise Exception("You must pass an ID to the Spreadsheet component as a dict with the keys type and id")
            if 'id' not in id:
                raise Exception("You must pass an ID to the Spreadsheet component as a dict with the keys type and id")
            if id['type'] != ID_TYPE:
                raise Exception("You must pass an ID to the Spreadsheet component as a dict with the keys type and id")
            
            # Get the current Dash app, and check that Mito has been activated on it
            from dash import get_app
            try:
                current_app = get_app()
            except:
                raise Exception("You must create a Dash app before using the Spreadsheet component. See https://docs.trymito.io/mito-for-dash/api-reference#activate_mito")
            if len(current_app.callback_map) == 0:
                raise Exception("You must activate Mito on your Dash app before using the Spreadsheet component. See https://docs.trymito.io/mito-for-dash/api-reference#activate_mito")

            self.id = id
            self.mito_id = id['id'] # TODO: document this
            # Note: num_messages must be ever increasing, so that whenever we get a new message, the spreadsheet_result
            # and spreadsheet_selection error strings change. This way, we can correctly trigger callbacks that correspond
            # to these values in all cases
            self.num_messages=0
            self.args = args
            self._set_new_mito_backend(
                *args, 
                session_key='',
                import_folder=import_folder, 
                code_options=code_options,
                df_names=df_names,
                sheet_functions=sheet_functions,
                importers=importers,
                editors=editors,
                theme=theme,
                mito_frontend_key=mito_frontend_key
            )

            self.track_selection = track_selection

            super(Spreadsheet, self).__init__()

            # We save the unprocessed messages in a list -- so that we can process them
            # in the callback in the order that they were received -- without them interrupting
            # eachother and having to deal with race conditions
            self.unprocessed_messages: Any = Queue()
            self.processing_messages = False

            self.index_and_selections: Optional[MitoFrontendIndexAndSelections] = None

            # Make sure to save import-folder and code-options as attributes, so if we need
            # to recreate the backend, we can do so
            self.import_folder = import_folder
            self.code_options = code_options
            self.df_names = df_names
            self.sheet_functions = sheet_functions
            self.importers = importers
            self.editors = editors
            self.theme = theme

            self.all_json = self.get_all_json()

            # Save the instance, so we can look it up later
            if self.mito_id not in self.__class__.instances:
                self.__class__.instances[self.mito_id] = (self, dict())

        @classmethod
        def get_instance(cls, mito_id: str, session_key: str) -> Optional[Any]:
            """
            The Spreadsheet component stores a map from the mito_id of the spreadsheet to a tuple of the 
            original Spreadsheet instances, as well as a map from session key to the session instance.

            This is because Mito is not properly stateless yet. As a result, to work around limitaitons of Dash, 
            we have the frontend keep track of the session key of the Mito spreadsheet, and then we duplicate the
            spreadsheet for each session key.

            This way every user gets a new and unique Mito backend. When Mito is stateless, we can remove this, 
            but it will require larger refactors to the Mito app.
            """

            # First, lookup the instance by mito_id
            instance = cls.instances.get(mito_id, None)

            if instance is None:
                return None

            session_instance = instance[1].get(session_key, None)
            if session_instance is not None:
                return session_instance
            
            # If we don't have a session_instance, then we need to create one
            # We do this by copying the instance, and then saving it
            session_instance = instance[0].safe_copy()
            cls.instances[mito_id][1][session_key] = session_instance
            return session_instance
        
        def safe_copy(self):
            """
            Make a safe copy of this Spreadsheet component. Should be called 
            on the original instance of the Spreadsheet component, and not on
            the session instance -- so the user starts with the original starting
            values.
            """

            return self.__class__(
                *self.args,
                id=self.id,
                import_folder=self.import_folder,
                code_options=self.code_options,
                df_names=self.df_names,
                sheet_functions=self.sheet_functions, 
                importers=self.importers,
                editors=self.editors,
                theme=self.theme,
                track_selection=self.track_selection,
                mito_frontend_key=self.mito_frontend_key
            )

        def _set_new_mito_backend(
                self, 
                *args: Union[pd.DataFrame, str, None], 
                session_key:Optional[str]=None,
                import_folder: Optional[str]=None,
                code_options: Optional[CodeOptions]=None,
                df_names: Optional[List[str]]=None,
                sheet_functions: Optional[List[Callable]]=None, 
                importers: Optional[List[Callable]]=None,
                editors: Optional[List[Callable]]=None,
                theme: Optional[MitoTheme]=None,
                mito_frontend_key: Optional[str]=None
            ) -> None:
            """
            Called when the component is created, or when the input data is changed.
            """
            self.mito_frontend_key = get_new_id() if mito_frontend_key is None else mito_frontend_key
            self.mito_backend = MitoBackend(
                *args, 
                import_folder=import_folder, 
                code_options=code_options,
                user_defined_functions=sheet_functions,
                user_defined_importers=importers,
                user_defined_editors=editors,
                theme=theme
            )
            self.responses: List[Dict[str, Any]] = []
            def send(response):
                self.responses.append(response)
            self.mito_backend.mito_send = send

            # If there are any df_names, then we send them to the backend as well. 
            # TODO: we should be able to pass this directly to the backend
            if df_names is not None and len(df_names) > 0:
                self.mito_backend.receive_message(
                    {
                        'event': 'update_event',
                        'id': get_new_id(),
                        'type': 'args_update',
                        'params': {
                            'args': df_names
                        },
                    }
                )

            # If you use the @callback decorator with spreadsheet_result or spreadsheet_selection, users will get these helpful error messages that 
            # will tell them what is wrong with their approach - namely, that they need to use `@mito_callback` instead. Note that this includes:
            # 1.    The num_messages so it triggers callbacks correct in the all cases
            # 2.    The prop name and the id of the spreadsheet so the `@mito_callback` function can inspect this string and replace this arg with the
            #       actual value the user really wants
            self.spreadsheet_result = WRONG_CALLBACK_ERROR_MESSAGE.format(prop_name='spreadsheet_result', num_messages=self.num_messages, id=self.mito_id, session_key=session_key)
            self.spreadsheet_selection = WRONG_CALLBACK_ERROR_MESSAGE.format(prop_name='spreadsheet_selection', num_messages=self.num_messages, id=self.mito_id, session_key=session_key)

                
        def process_single_message(self, session_key: str) -> None:

            # If we are already processing messages -- then wait until it is
            if self.processing_messages:
                while self.processing_messages:
                    time.sleep(0.1)
                
            # Otherwise, set the processing flag to true
            self.processing_messages = True

            # Process all the messages in the queue
            try:
                if not self.unprocessed_messages.empty():
                    value = self.unprocessed_messages.get()
                    self.mito_backend.receive_message(value)
            except:
                # Make sure we always set the processing flag to false
                pass
                
            # Set the processing flag to false
            self.processing_messages = False

            self.spreadsheet_result = WRONG_CALLBACK_ERROR_MESSAGE.format(prop_name='spreadsheet_result', num_messages=self.num_messages, id=self.mito_id, session_key=session_key)
            
            
        def get_all_json(self) -> str:
            return json.dumps({
                **self.mito_backend.get_shared_state_variables(),
                'responses_json': json.dumps(self.responses),
                'key': self.mito_frontend_key,
                'track_selection': self.track_selection
            })
        
        def get_result(self):
            return SpreadsheetResult(
                dfs=[df.copy() for df in self.mito_backend.steps_manager.dfs],
                code=self.mito_backend.steps_manager.code(),
                index_and_selections=self.index_and_selections,
                fully_parameterized_function=self.mito_backend.fully_parameterized_function,
                param_metadata=self.mito_backend.param_metadata,
                code_options=self.code_options
            )
        
except ImportError:

    class Spreadsheet(): # type: ignore
        def __init__(self, *args, **kwargs):
            raise Exception("You must install dash to use the Spreadsheet component")